#
# Copyright (c) 2000 Tyler C. Sarna <tsarna@sarna.org>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#      This product includes software developed by Tyler C. Sarna.
# 4. Neither the name of the author nor the names of contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

from reportlab.lib.units import inch
from reportlab.lib.utils import asNative
from reportlab.graphics.barcode.common import MultiWidthBarcode
from string import digits

_patterns = {
    0   :   'BaBbBb',    1   :   'BbBaBb',    2   :   'BbBbBa',
    3   :   'AbAbBc',    4   :   'AbAcBb',    5   :   'AcAbBb',
    6   :   'AbBbAc',    7   :   'AbBcAb',    8   :   'AcBbAb',
    9   :   'BbAbAc',    10  :   'BbAcAb',    11  :   'BcAbAb',
    12  :   'AaBbCb',    13  :   'AbBaCb',    14  :   'AbBbCa',
    15  :   'AaCbBb',    16  :   'AbCaBb',    17  :   'AbCbBa',
    18  :   'BbCbAa',    19  :   'BbAaCb',    20  :   'BbAbCa',
    21  :   'BaCbAb',    22  :   'BbCaAb',    23  :   'CaBaCa',
    24  :   'CaAbBb',    25  :   'CbAaBb',    26  :   'CbAbBa',
    27  :   'CaBbAb',    28  :   'CbBaAb',    29  :   'CbBbAa',
    30  :   'BaBaBc',    31  :   'BaBcBa',    32  :   'BcBaBa',
    33  :   'AaAcBc',    34  :   'AcAaBc',    35  :   'AcAcBa',
    36  :   'AaBcAc',    37  :   'AcBaAc',    38  :   'AcBcAa',
    39  :   'BaAcAc',    40  :   'BcAaAc',    41  :   'BcAcAa',
    42  :   'AaBaCc',    43  :   'AaBcCa',    44  :   'AcBaCa',
    45  :   'AaCaBc',    46  :   'AaCcBa',    47  :   'AcCaBa',
    48  :   'CaCaBa',    49  :   'BaAcCa',    50  :   'BcAaCa',
    51  :   'BaCaAc',    52  :   'BaCcAa',    53  :   'BaCaCa',
    54  :   'CaAaBc',    55  :   'CaAcBa',    56  :   'CcAaBa',
    57  :   'CaBaAc',    58  :   'CaBcAa',    59  :   'CcBaAa',
    60  :   'CaDaAa',    61  :   'BbAdAa',    62  :   'DcAaAa',
    63  :   'AaAbBd',    64  :   'AaAdBb',    65  :   'AbAaBd',
    66  :   'AbAdBa',    67  :   'AdAaBb',    68  :   'AdAbBa',
    69  :   'AaBbAd',    70  :   'AaBdAb',    71  :   'AbBaAd',
    72  :   'AbBdAa',    73  :   'AdBaAb',    74  :   'AdBbAa',
    75  :   'BdAbAa',    76  :   'BbAaAd',    77  :   'DaCaAa',
    78  :   'BdAaAb',    79  :   'AcDaAa',    80  :   'AaAbDb',
    81  :   'AbAaDb',    82  :   'AbAbDa',    83  :   'AaDbAb',
    84  :   'AbDaAb',    85  :   'AbDbAa',    86  :   'DaAbAb',
    87  :   'DbAaAb',    88  :   'DbAbAa',    89  :   'BaBaDa',
    90  :   'BaDaBa',    91  :   'DaBaBa',    92  :   'AaAaDc',
    93  :   'AaAcDa',    94  :   'AcAaDa',    95  :   'AaDaAc',
    96  :   'AaDcAa',    97  :   'DaAaAc',    98  :   'DaAcAa',
    99  :   'AaCaDa',    100 :   'AaDaCa',    101 :   'CaAaDa',
    102 :   'DaAaCa',    103 :   'BaAdAb',    104 :   'BaAbAd',
    105 :   'BaAbCb',    106 :   'BcCaAaB'
}

starta, startb, startc, stop = 103, 104, 105, 106

seta = {
        ' ' :   0,        '!' :   1,        '"' :   2,        '#' :   3,
        '$' :   4,        '%' :   5,        '&' :   6,       '\'' :   7,
        '(' :   8,        ')' :   9,        '*' :  10,        '+' :  11,
        ',' :  12,        '-' :  13,        '.' :  14,        '/' :  15,
        '0' :  16,        '1' :  17,        '2' :  18,        '3' :  19,
        '4' :  20,        '5' :  21,        '6' :  22,        '7' :  23,
        '8' :  24,        '9' :  25,        ':' :  26,        ';' :  27,
        '<' :  28,        '=' :  29,        '>' :  30,        '?' :  31,
        '@' :  32,        'A' :  33,        'B' :  34,        'C' :  35,
        'D' :  36,        'E' :  37,        'F' :  38,        'G' :  39,
        'H' :  40,        'I' :  41,        'J' :  42,        'K' :  43,
        'L' :  44,        'M' :  45,        'N' :  46,        'O' :  47,
        'P' :  48,        'Q' :  49,        'R' :  50,        'S' :  51,
        'T' :  52,        'U' :  53,        'V' :  54,        'W' :  55,
        'X' :  56,        'Y' :  57,        'Z' :  58,        '[' :  59,
       '\\' :  60,        ']' :  61,        '^' :  62,        '_' :  63,
     '\x00' :  64,     '\x01' :  65,     '\x02' :  66,     '\x03' :  67,
     '\x04' :  68,     '\x05' :  69,     '\x06' :  70,     '\x07' :  71,
     '\x08' :  72,     '\x09' :  73,     '\x0a' :  74,     '\x0b' :  75,
     '\x0c' :  76,     '\x0d' :  77,     '\x0e' :  78,     '\x0f' :  79,
     '\x10' :  80,     '\x11' :  81,     '\x12' :  82,     '\x13' :  83,
     '\x14' :  84,     '\x15' :  85,     '\x16' :  86,     '\x17' :  87,
     '\x18' :  88,     '\x19' :  89,     '\x1a' :  90,     '\x1b' :  91,
     '\x1c' :  92,     '\x1d' :  93,     '\x1e' :  94,     '\x1f' :  95,
     '\xf3' :  96,     '\xf2' :  97,    'SHIFT' :  98,     'TO_C' :  99,
     'TO_B' : 100,     '\xf4' : 101,     '\xf1' : 102
}

setb = {
        ' ' :   0,        '!' :   1,        '"' :   2,        '#' :   3,
        '$' :   4,        '%' :   5,        '&' :   6,       '\'' :   7,
        '(' :   8,        ')' :   9,        '*' :  10,        '+' :  11,
        ',' :  12,        '-' :  13,        '.' :  14,        '/' :  15,
        '0' :  16,        '1' :  17,        '2' :  18,        '3' :  19,
        '4' :  20,        '5' :  21,        '6' :  22,        '7' :  23,
        '8' :  24,        '9' :  25,        ':' :  26,        ';' :  27,
        '<' :  28,        '=' :  29,        '>' :  30,        '?' :  31,
        '@' :  32,        'A' :  33,        'B' :  34,        'C' :  35,
        'D' :  36,        'E' :  37,        'F' :  38,        'G' :  39,
        'H' :  40,        'I' :  41,        'J' :  42,        'K' :  43,
        'L' :  44,        'M' :  45,        'N' :  46,        'O' :  47,
        'P' :  48,        'Q' :  49,        'R' :  50,        'S' :  51,
        'T' :  52,        'U' :  53,        'V' :  54,        'W' :  55,
        'X' :  56,        'Y' :  57,        'Z' :  58,        '[' :  59,
       '\\' :  60,        ']' :  61,        '^' :  62,        '_' :  63,
        '`' :  64,        'a' :  65,        'b' :  66,        'c' :  67,
        'd' :  68,        'e' :  69,        'f' :  70,        'g' :  71,
        'h' :  72,        'i' :  73,        'j' :  74,        'k' :  75,
        'l' :  76,        'm' :  77,        'n' :  78,        'o' :  79,
        'p' :  80,        'q' :  81,        'r' :  82,        's' :  83,
        't' :  84,        'u' :  85,        'v' :  86,        'w' :  87,
        'x' :  88,        'y' :  89,        'z' :  90,        '{' :  91,
        '|' :  92,        '}' :  93,        '~' :  94,     '\x7f' :  95,
     '\xf3' :  96,     '\xf2' :  97,    'SHIFT' :  98,     'TO_C' :  99,
     '\xf4' : 100,     'TO_A' : 101,     '\xf1' : 102
}

setc = {
    '00': 0, '01': 1, '02': 2, '03': 3, '04': 4,
    '05': 5, '06': 6, '07': 7, '08': 8, '09': 9,
    '10':10, '11':11, '12':12, '13':13, '14':14,
    '15':15, '16':16, '17':17, '18':18, '19':19,
    '20':20, '21':21, '22':22, '23':23, '24':24,
    '25':25, '26':26, '27':27, '28':28, '29':29,
    '30':30, '31':31, '32':32, '33':33, '34':34,
    '35':35, '36':36, '37':37, '38':38, '39':39,
    '40':40, '41':41, '42':42, '43':43, '44':44,
    '45':45, '46':46, '47':47, '48':48, '49':49,
    '50':50, '51':51, '52':52, '53':53, '54':54,
    '55':55, '56':56, '57':57, '58':58, '59':59,
    '60':60, '61':61, '62':62, '63':63, '64':64,
    '65':65, '66':66, '67':67, '68':68, '69':69,
    '70':70, '71':71, '72':72, '73':73, '74':74,
    '75':75, '76':76, '77':77, '78':78, '79':79,
    '80':80, '81':81, '82':82, '83':83, '84':84,
    '85':85, '86':86, '87':87, '88':88, '89':89,
    '90':90, '91':91, '92':92, '93':93, '94':94,
    '95':95, '96':96, '97':97, '98':98, '99':99,

    'TO_B' : 100,    'TO_A' : 101,    '\xf1' : 102
}

setmap = {
    'TO_A' : (seta, setb),
    'TO_B' : (setb, seta),
    'TO_C' : (setc, None),
    'START_A' : (starta, seta, setb),
    'START_B' : (startb, setb, seta),
    'START_C' : (startc, setc, None),
}
cStarts = ('START_B','TO_A','TO_B')
tos = list(setmap.keys())

class Code128(MultiWidthBarcode):
    """
    Code 128 is a very compact symbology that can encode the entire
    128 character ASCII set, plus 4 special control codes,
    (FNC1-FNC4, expressed in the input string as \xf1 to \xf4).
    Code 128 can also encode digits at double density (2 per byte)
    and has a mandatory checksum.  Code 128 is well supported and
    commonly used -- for example, by UPS for tracking labels.
    
    Because of these qualities, Code 128 is probably the best choice
    for a linear symbology today (assuming you have a choice).

    Options that may be passed to constructor:

        value (int, or numeric string. required.):
            The value to encode.
   
        barWidth (float, default .0075):
            X-Dimension, or width of the smallest element
            Minumum is .0075 inch (7.5 mils).
            
        barHeight (float, see default below):
            Height of the symbol.  Default is the height of the two
            bearer bars (if they exist) plus the greater of .25 inch
            or .15 times the symbol's length.

        quiet (bool, default 1):
            Wether to include quiet zones in the symbol.
            
        lquiet (float, see default below):
            Quiet zone size to left of code, if quiet is true.
            Default is the greater of .25 inch, or 10 barWidth
            
        rquiet (float, defaults as above):
            Quiet zone size to right left of code, if quiet is true.
            
    Sources of Information on Code 128:

    http://www.semiconductor.agilent.com/barcode/sg/Misc/code_128.html
    http://www.adams1.com/pub/russadam/128code.html
    http://www.barcodeman.com/c128.html

    Official Spec, "ANSI/AIM BC4-1999, ISS" is available for US$45 from
    http://www.aimglobal.org/aimstore/
    """
    barWidth = inch * 0.0075
    lquiet = None
    rquiet = None
    quiet = 1
    barHeight = None
    def __init__(self, value='', **args):
        value = str(value) if isinstance(value,int) else asNative(value)
            
        for k, v in args.items():
            setattr(self, k, v)

        if self.quiet:
            if self.lquiet is None:
                self.lquiet = max(inch * 0.25, self.barWidth * 10.0)
            if self.rquiet is None:
                self.rquiet = max(inch * 0.25, self.barWidth * 10.0)
        else:
            self.lquiet = self.rquiet = 0.0

        MultiWidthBarcode.__init__(self, value)

    def validate(self):
        vval = ""
        self.valid = 1
        for c in self.value:
            if ord(c) > 127 and c not in '\xf1\xf2\xf3\xf4':
                self.valid = 0
                continue
            vval = vval + c
        self.validated = vval
        return vval


    def _try_TO_C(self, l):
        '''Improved version of old _trailingDigitsToC(self, l) inspired by'''
        i = 0
        nl = []
        while i < len(l):
            startpos = i
            rl = []
            savings = -1 # the TO_C costs one character
            while i < len(l):
                if l[i] in cStarts:
                    j = i
                    break
                elif l[i] == '\xf1':
                    rl.append(l[i])
                    i += 1
                    continue
                elif l[i] in digits \
                    and l[i+1] in digits:
                    rl.append(l[i] + l[i+1])
                    i += 2
                    savings += 1
                    continue
                else:
                    if l[i] in digits and l[i+1]=='STOP':
                        rrl = []
                        rsavings = -1   #we need a TO_C
                        k = i
                        while k>startpos:
                            if l[k]=='\xf1':
                                rrl.append(l[i])
                                k -= 1
                            elif l[k] in digits and l[k-1] in digits:
                                rrl.append(l[k-1]+l[k])
                                rsavings += 1
                                k -= 2
                            else:
                                break
                        rrl.reverse()
                        if rsavings>savings+int(savings>=0 and (startpos and nl[-1] in cStarts))-1:
                            nl += l[startpos]
                            startpos += 1
                            rl = rrl
                            del rrl
                            i += 1
                    break
            ta = not (l[i]=='STOP' or j==i)
            xs = savings>=0 and (startpos and nl[-1] in cStarts)
            if savings+int(xs) > int(ta):
                if xs:
                    toc = nl[-1][:-1]+'C'
                    del nl[-1]
                else:
                    toc = 'TO_C'
                nl += [toc]+rl
                if ta:
                    nl.append('TO'+l[j][-2:])
                nl.append(l[i])
            else:
                nl += l[startpos:i+1]
            i += 1
        return nl

    def encode(self):
        # First, encode using only B
        s = self.validated
        l = ['START_B']
        for c in s:
            if c not in setb:
                l = l + ['TO_A', c, 'TO_B']
            else:
                l.append(c)
        l.append('STOP')

        l = self._try_TO_C(l)

        # Finally, replace START_X,TO_Y with START_Y
        if l[1] in tos:
            l[:2] = ['START_' + l[1][-1]]

#        print repr(l)

        # encode into numbers
        start, set, shset = setmap[l[0]]
        e = [start]
        
        l = l[1:-1]
        while l:
            c = l[0]
            if c == 'SHIFT':
                e = e + [set[c], shset[l[1]]]
                l = l[2:]
            elif c in tos:
                e.append(set[c])
                set, shset = setmap[c]
                l = l[1:]
            else:
                e.append(set[c])
                l = l[1:]

        c = e[0]
        for i in range(1, len(e)):
            c = c + i * e[i]
        self.encoded = e + [c % 103, stop]
        return self.encoded

    def decompose(self):
        self.decomposed = ''.join([_patterns[c] for c in self.encoded])
        return self.decomposed

    def _humanText(self):
        return self.value

class Code128Auto(Code128):
    '''contributed by https://bitbucket.org/kylemacfarlane/
    see https://bitbucket.org/rptlab/reportlab/issues/69/implementations-of-code-128-auto-and-data
    '''
    def encode(self):
        s = self.validated

        current_set = None
        l = []
        value = list(s)
        while value:
            c = value.pop(0)
            if c in digits and value and value[0] in digits:
                c += value.pop(0)

            if c in setc:
                set_ = 'C'
            elif c in setb:
                set_ = 'B'
            else:
                set_ = 'A'

            if current_set != set_:
                if current_set:
                    l.append('TO_' + set_)
                else:
                    l.append('START_' + set_)
                current_set = set_

            l.append(c)
        l.append('STOP')

        start, set, shset = setmap[l[0]]
        e = [start]

        l = l[1:-1]
        while l:
            c = l[0]
            if c == 'SHIFT':
                e = e + [set[c], shset[l[1]]]
                l = l[2:]
            elif c in tos:
                e.append(set[c])
                set, shset = setmap[c]
                l = l[1:]
            else:
                e.append(set[c])
                l = l[1:]

        c = e[0]
        for i in range(1, len(e)):
            c = c + i * e[i]
        self.encoded = e + [c % 103, stop]
        return self.encoded

if __name__=='__main__':
    def main():
        from reportlab.graphics.barcode.code128 import Code128
        from reportlab.platypus import Spacer, SimpleDocTemplate
        from reportlab.lib.units import inch
        from reportlab.lib.styles import getSampleStyleSheet
        from reportlab.platypus.paragraph import Paragraph
        from reportlab.platypus.flowables import KeepTogether
        styles = getSampleStyleSheet()
        styleN = styles['Normal']
        styleH = styles['Heading1']
        story = []
        storyAdd = story.append
        for s in (
            'BBBB123456BBB',
            'BBBB12345BBB',
            'BBBB1234BBB',
            'BBBB123BBB',
            'BBBB12BBB',
            'BBBB1BBB',
            'BBBB123456aa',
            'BBBB1234aa',
            'BBBB123aa',
            'BBBB12aa',
            'BBBB1aa',
            'BBBB123456',
            'BBBB12345',
            'BBBB1234',
            'BBBB123',
            'BBBB12',
            'BBBB1',
            '\xf11234B',
            'Ba\xf11234B',
            'Ba12',
            'Ba123B',
            'Ba1234B',
            'BBBB1234567',
            'BBBB1234567aa',
            ):
            storyAdd(KeepTogether([Paragraph('Code 128 %r' % s, styleN),Code128(s)]))
            storyAdd(Spacer(inch,inch))
        SimpleDocTemplate('code128-out.pdf').build(story)
    main()



# (c) 2008 Jerome Alet - <alet@librelogiciel.com>
# Licensing terms : ReportLab's license.

from reportlab.graphics.barcode.code39 import Standard39
from reportlab.lib import colors
from reportlab.lib.units import cm
from string import ascii_uppercase, digits as string_digits

class BaseLTOLabel(Standard39) :
    """
    Base class for LTO labels.

    Specification taken from "IBM LTO Ultrium Cartridge Label Specification, Revision 3"
    available on  May 14th 2008 from :
    http://www-1.ibm.com/support/docview.wss?rs=543&context=STCVQ6R&q1=ssg1*&uid=ssg1S7000429&loc=en_US&cs=utf-8&lang=en+en
    """
    LABELWIDTH = 7.9 * cm
    LABELHEIGHT = 1.7 * cm
    LABELROUND = 0.15 * cm
    CODERATIO = 2.75
    CODENOMINALWIDTH = 7.4088 * cm
    CODEBARHEIGHT = 1.11 * cm
    CODEBARWIDTH = 0.0432 * cm
    CODEGAP = CODEBARWIDTH
    CODELQUIET = 10 * CODEBARWIDTH
    CODERQUIET = 10 * CODEBARWIDTH
    def __init__(self, prefix="",
                       number=None,
                       subtype="1",
                       border=None,
                       checksum=False,
                       availheight=None) :
        """
           Initializes an LTO label.

           prefix : Up to six characters from [A-Z][0-9]. Defaults to "".
           number : Label's number or None. Defaults to None.
           subtype : LTO subtype string , e.g. "1" for LTO1. Defaults to "1".
           border : None, or the width of the label's border. Defaults to None.
           checksum : Boolean indicates if checksum char has to be printed. Defaults to False.
           availheight : Available height on the label, or None for automatic. Defaults to None.
        """
        self.height = max(availheight, self.CODEBARHEIGHT)
        self.border = border
        if (len(subtype) != 1) \
            or (subtype not in ascii_uppercase + string_digits) :
            raise ValueError("Invalid subtype '%s'" % subtype)
        if ((not number) and (len(prefix) > 6)) \
           or not prefix.isalnum() :
            raise ValueError("Invalid prefix '%s'" % prefix)
        label = "%sL%s" % ((prefix + str(number or 0).zfill(6 - len(prefix)))[:6],
                           subtype)
        if len(label) != 8 :
            raise ValueError("Invalid set of parameters (%s, %s, %s)" \
                                % (prefix, number, subtype))
        self.label = label
        Standard39.__init__(self,
                            label,
                            ratio=self.CODERATIO,
                            barHeight=self.height,
                            barWidth=self.CODEBARWIDTH,
                            gap=self.CODEGAP,
                            lquiet=self.CODELQUIET,
                            rquiet=self.CODERQUIET,
                            quiet=True,
                            checksum=checksum)

    def drawOn(self, canvas, x, y) :
        """Draws the LTO label onto the canvas."""
        canvas.saveState()
        canvas.translate(x, y)
        if self.border :
            canvas.setLineWidth(self.border)
            canvas.roundRect(0, 0,
                        self.LABELWIDTH,
                        self.LABELHEIGHT,
                        self.LABELROUND)
        Standard39.drawOn(self,
                          canvas,
                          (self.LABELWIDTH-self.CODENOMINALWIDTH)/2.0,
                          self.LABELHEIGHT-self.height)
        canvas.restoreState()

class VerticalLTOLabel(BaseLTOLabel) :
    """
    A class for LTO labels with rectangular blocks around the tape identifier.
    """
    LABELFONT = ("Helvetica-Bold", 14)
    BLOCKWIDTH = 1*cm
    BLOCKHEIGHT = 0.45*cm
    LINEWIDTH = 0.0125
    NBBLOCKS = 7
    COLORSCHEME = ("red",
                   "yellow",
                   "lightgreen",
                   "lightblue",
                   "grey",
                   "orangered",
                   "pink",
                   "darkgreen",
                   "orange",
                   "purple")

    def __init__(self, *args, **kwargs) :
        """
        Initializes the label.

        colored : boolean to determine if blocks have to be colorized.
        """
        if "colored" in kwargs:
            self.colored = kwargs["colored"]
            del kwargs["colored"]
        else :
            self.colored = False
        kwargs["availheight"] = self.LABELHEIGHT-self.BLOCKHEIGHT
        BaseLTOLabel.__init__(self, *args, **kwargs)

    def drawOn(self, canvas, x, y) :
        """Draws some blocks around the identifier's characters."""
        BaseLTOLabel.drawOn(self,
                            canvas,
                            x,
                            y)
        canvas.saveState()
        canvas.setLineWidth(self.LINEWIDTH)
        canvas.setStrokeColorRGB(0, 0, 0)
        canvas.translate(x, y)
        xblocks = (self.LABELWIDTH-(self.NBBLOCKS*self.BLOCKWIDTH))/2.0
        for i in range(self.NBBLOCKS) :
            (font, size) = self.LABELFONT
            newfont = self.LABELFONT
            if i == (self.NBBLOCKS - 1) :
                part = self.label[i:]
                (font, size) = newfont
                size /= 2.0
                newfont = (font, size)
            else :
                part = self.label[i]
            canvas.saveState()
            canvas.translate(xblocks+(i*self.BLOCKWIDTH), 0)
            if self.colored and part.isdigit() :
                canvas.setFillColorRGB(*getattr(colors,
                                                self.COLORSCHEME[int(part)],
                                                colors.Color(1, 1, 1)).rgb())
            else:
                canvas.setFillColorRGB(1, 1, 1)
            canvas.rect(0, 0, self.BLOCKWIDTH, self.BLOCKHEIGHT, fill=True)
            canvas.translate((self.BLOCKWIDTH+canvas.stringWidth(part, *newfont))/2.0,
                             (self.BLOCKHEIGHT/2.0))
            canvas.rotate(90.0)
            canvas.setFont(*newfont)
            canvas.setFillColorRGB(0, 0, 0)
            canvas.drawCentredString(0, 0, part)
            canvas.restoreState()
        canvas.restoreState()

def test() :
    """Test this."""
    from reportlab.pdfgen.canvas import Canvas
    from reportlab.lib import pagesizes

    canvas = Canvas("labels.pdf", pagesize=pagesizes.A4)
    canvas.setFont("Helvetica", 30)
    (width, height) = pagesizes.A4
    canvas.drawCentredString(width/2.0, height-4*cm, "Sample LTO labels")
    xpos = xorig = 2 * cm
    ypos = yorig = 2 * cm
    colwidth = 10 * cm
    lineheight = 3.9 * cm
    count = 1234
    BaseLTOLabel("RL", count, "3").drawOn(canvas, xpos, ypos)
    ypos += lineheight
    count += 1
    BaseLTOLabel("RL", count, "3",
                 border=0.0125).drawOn(canvas, xpos, ypos)
    ypos += lineheight
    count += 1
    VerticalLTOLabel("RL", count, "3").drawOn(canvas, xpos, ypos)
    ypos += lineheight
    count += 1
    VerticalLTOLabel("RL", count, "3",
                    border=0.0125).drawOn(canvas, xpos, ypos)
    ypos += lineheight
    count += 1
    VerticalLTOLabel("RL", count, "3",
                    colored=True).drawOn(canvas, xpos, ypos)
    ypos += lineheight
    count += 1
    VerticalLTOLabel("RL", count, "3",
                    border=0.0125, colored=True).drawOn(canvas, xpos, ypos)
    canvas.showPage()
    canvas.save()

if __name__ == "__main__" :
    test()



#Copyright ReportLab Europe Ltd. 2000-2021
#see license.txt for license details
#history https://hg.reportlab.com/hg-public/reportlab/log/tip/src/reportlab/graphics/renderbase.py

__version__='3.13.0'
__doc__='''Superclass for renderers to factor out common functionality and default implementations.'''

from reportlab.graphics.shapes import *
from reportlab.lib.validators import DerivedValue
from reportlab import rl_config

from . transform import mmult, inverse

def getStateDelta(shape):
    """Used to compute when we need to change the graphics state.
    For example, if we have two adjacent red shapes we don't need
    to set the pen color to red in between. Returns the effect
    the given shape would have on the graphics state"""
    delta = {}
    for prop, value in shape.getProperties().items():
        if prop in STATE_DEFAULTS:
            delta[prop] = value
    return delta

class StateTracker:
    """Keeps a stack of transforms and state
    properties.  It can contain any properties you
    want, but the keys 'transform' and 'ctm' have
    special meanings.  The getCTM()
    method returns the current transformation
    matrix at any point, without needing to
    invert matrixes when you pop."""
    def __init__(self, defaults=None, defaultObj=None):
        # one stack to keep track of what changes...
        self._deltas = []

        # and another to keep track of cumulative effects.  Last one in
        # list is the current graphics state.  We put one in to simplify
        # loops below.
        self._combined = []
        if defaults is None:
            defaults = STATE_DEFAULTS.copy()
        if defaultObj:
            for k in STATE_DEFAULTS.keys():
                a = 'initial'+k[:1].upper()+k[1:]
                if hasattr(defaultObj,a):
                    defaults[k] = getattr(defaultObj,a)
        #ensure  that if we have a transform, we have a CTM
        if 'transform' in defaults:
            defaults['ctm'] = defaults['transform']
        self._combined.append(defaults)

    def _applyDefaultObj(self,d):
        return d

    def push(self,delta):
        """Take a new state dictionary of changes and push it onto
        the stack.  After doing this, the combined state is accessible
        through getState()"""

        newstate = self._combined[-1].copy()
        for key, value in delta.items():
            if key == 'transform':  #do cumulative matrix
                newstate['transform'] = delta['transform']
                newstate['ctm'] = mmult(self._combined[-1]['ctm'], delta['transform'])
                #print 'statetracker transform = (%0.2f, %0.2f, %0.2f, %0.2f, %0.2f, %0.2f)' % tuple(newstate['transform'])
                #print 'statetracker ctm = (%0.2f, %0.2f, %0.2f, %0.2f, %0.2f, %0.2f)' % tuple(newstate['ctm'])

            else:  #just overwrite it
                newstate[key] = value

        self._combined.append(newstate)
        self._deltas.append(delta)

    def pop(self):
        """steps back one, and returns a state dictionary with the
        deltas to reverse out of wherever you are.  Depending
        on your back end, you may not need the return value,
        since you can get the complete state afterwards with getState()"""
        del self._combined[-1]
        newState = self._combined[-1]
        lastDelta = self._deltas[-1]
        del  self._deltas[-1]
        #need to diff this against the last one in the state
        reverseDelta = {}
        #print 'pop()...'
        for key, curValue in lastDelta.items():
            #print '   key=%s, value=%s' % (key, curValue)
            prevValue = newState[key]
            if prevValue != curValue:
                #print '    state popping "%s"="%s"' % (key, curValue)
                if key == 'transform':
                    reverseDelta[key] = inverse(lastDelta['transform'])
                else:  #just return to previous state
                    reverseDelta[key] = prevValue
        return reverseDelta

    def getState(self):
        "returns the complete graphics state at this point"
        return self._combined[-1]

    def getCTM(self):
        "returns the current transformation matrix at this point"""
        return self._combined[-1]['ctm']

    def __getitem__(self,key):
        "returns the complete graphics state value of key at this point"
        return self._combined[-1][key]

    def __setitem__(self,key,value):
        "sets the complete graphics state value of key to value"
        self._combined[-1][key] = value

def testStateTracker():
    print('Testing state tracker')
    defaults = {'fillColor':None, 'strokeColor':None,'fontName':None, 'transform':[1,0,0,1,0,0]}
    from reportlab.graphics.shapes import _baseGFontName
    deltas = [
        {'fillColor':'red'},
        {'fillColor':'green', 'strokeColor':'blue','fontName':_baseGFontName},
        {'transform':[0.5,0,0,0.5,0,0]},
        {'transform':[0.5,0,0,0.5,2,3]},
        {'strokeColor':'red'}
        ]

    st = StateTracker(defaults)
    print('initial:', st.getState())
    print()
    for delta in deltas:
        print('pushing:', delta)
        st.push(delta)
        print('state:  ',st.getState(),'\n')

    for delta in deltas:
        print('popping:',st.pop())
        print('state:  ',st.getState(),'\n')

def _expandUserNode(node,canvas):
    if isinstance(node, UserNode):
        try:
            if hasattr(node,'_canvas'):
                ocanvas = 1
            else:
                node._canvas = canvas
                ocanvas = None
            onode = node
            node = node.provideNode()
        finally:
            if not ocanvas: del onode._canvas
    return node

def renderScaledDrawing(d):
    renderScale = d.renderScale
    if renderScale!=1.0:
        o = d
        d = d.__class__(o.width*renderScale,o.height*renderScale)
        d.__dict__ = o.__dict__.copy()
        d.scale(renderScale,renderScale)
        d.renderScale = 1.0
    return d

class Renderer:
    """Virtual superclass for graphics renderers."""

    def undefined(self, operation):
        raise ValueError("%s operation not defined at superclass class=%s" %(operation, self.__class__))

    def draw(self, drawing, canvas, x=0, y=0, showBoundary=rl_config._unset_):
        """This is the top level function, which draws the drawing at the given
        location. The recursive part is handled by drawNode."""
        self._tracker = StateTracker(defaultObj=drawing)
        #stash references for ease of  communication
        if showBoundary is rl_config._unset_: showBoundary=rl_config.showBoundary
        self._canvas = canvas
        canvas.__dict__['_drawing'] = self._drawing = drawing
        drawing._parent = None
        try:
            #bounding box
            if showBoundary:
                if hasattr(canvas,'drawBoundary'):
                    canvas.drawBoundary(showBoundary,x,y,drawing.width,drawing.height)
                else:
                    canvas.rect(x, y, drawing.width, drawing.height)
            canvas.saveState()
            self.initState(x,y)  #this is the push()
            self.drawNode(drawing)
            self.pop()
            canvas.restoreState()
        finally:
            #remove any circular references
            del self._canvas, self._drawing, canvas._drawing, drawing._parent, self._tracker

    def initState(self,x,y):
        deltas = self._tracker._combined[-1]
        deltas['transform'] = tuple(list(deltas['transform'])[:4])+(x,y)
        self._tracker.push(deltas)
        self.applyStateChanges(deltas, {})

    def pop(self):
        self._tracker.pop()

    def drawNode(self, node):
        """This is the recursive method called for each node
        in the tree"""
        # Undefined here, but with closer analysis probably can be handled in superclass
        self.undefined("drawNode")

    def getStateValue(self, key):
        """Return current state parameter for given key"""
        currentState = self._tracker._combined[-1]
        return currentState[key]

    def fillDerivedValues(self, node):
        """Examine a node for any values which are Derived,
        and replace them with their calculated values.
        Generally things may look at the drawing or their
        parent.

        """
        for key, value in node.__dict__.items():
            if isinstance(value, DerivedValue):
                #just replace with default for key?
                #print '    fillDerivedValues(%s)' % key
                newValue = value.getValue(self, key)
                #print '   got value of %s' % newValue
                node.__dict__[key] = newValue

    def drawNodeDispatcher(self, anode):
        """dispatch on the node's (super) class: shared code"""
        canvas = getattr(self,'_canvas',None)

        try:
            # replace UserNode with its contents
            node = _expandUserNode(anode,canvas)
            if not node: return
            if hasattr(node,'_canvas'):
                ocanvas = 1
            else:
                node._canvas = canvas
                ocanvas = None
            nodeparent = node is not anode and not hasattr(node,'_parent')
            if nodeparent: node._parent = anode

            self.fillDerivedValues(node)
            dtcb = getattr(node,'_drawTimeCallback',None)
            if dtcb:
                dtcb(node,canvas=canvas,renderer=self)
            #draw the object, or recurse
            if isinstance(node, Line):
                self.drawLine(node)
            elif isinstance(node, Path):
                self.drawPath(node)
            elif isinstance(node, String):
                self.drawString(node)
            elif isinstance(node, Group):
                self.drawGroup(node)
            elif isinstance(node, Rect):
                self.drawRect(node)
            elif isinstance(node, Image):
                self.drawImage(node)
            elif isinstance(node, Circle):
                self.drawCircle(node)
            elif isinstance(node, Ellipse):
                self.drawEllipse(node)
            elif isinstance(node, PolyLine):
                self.drawPolyLine(node)
            elif isinstance(node, Polygon):
                self.drawPolygon(node)
            elif isinstance(node, Wedge):
                self.drawWedge(node)
            elif isinstance(node, DirectDraw):
                node.drawDirectly(self)
            else:
                print('DrawingError','Unexpected element %s in drawing!' % str(node))
        finally:
            if not ocanvas: del node._canvas
            if nodeparent: del node._parent

    _restores = {'stroke':'_stroke','stroke_width': '_lineWidth','stroke_linecap':'_lineCap',
                'stroke_linejoin':'_lineJoin','fill':'_fill','font_family':'_font',
                'font_size':'_fontSize'}

    def drawGroup(self, group):
        # just do the contents.  Some renderers might need to override this
        # if they need a flipped transform
        canvas = getattr(self,'_canvas',None)
        for anode in group.getContents():
            node = _expandUserNode(anode,canvas)
            if not node: continue

            #here is where we do derived values - this seems to get everything. Touch wood.
            self.fillDerivedValues(node)
            try:
                if hasattr(node,'_canvas'):
                    ocanvas = 1
                else:
                    node._canvas = canvas
                    ocanvas = None
                if node is not anode:
                    anode._parent = group
                    node._parent = anode
                else:
                    node._parent = group
                self.drawNode(node)
            finally:
                if node is not anode: del anode._parent
                del node._parent
                if not ocanvas: del node._canvas

    def drawWedge(self, wedge):
        # by default ask the wedge to make a polygon of itself and draw that!
        #print "drawWedge"
        P = wedge.asPolygon()
        if isinstance(P,Path):
            self.drawPath(P)
        else:
            self.drawPolygon(P)

    def drawPath(self, path):
        polygons = path.asPolygons()
        for polygon in polygons:
                self.drawPolygon(polygon)

    def drawRect(self, rect):
        # could be implemented in terms of polygon
        self.undefined("drawRect")

    def drawLine(self, line):
        self.undefined("drawLine")

    def drawCircle(self, circle):
        self.undefined("drawCircle")

    def drawPolyLine(self, p):
        self.undefined("drawPolyLine")

    def drawEllipse(self, ellipse):
        self.undefined("drawEllipse")

    def drawPolygon(self, p):
        self.undefined("drawPolygon")

    def drawString(self, stringObj):
        self.undefined("drawString")

    def applyStateChanges(self, delta, newState):
        """This takes a set of states, and outputs the operators
        needed to set those properties"""
        self.undefined("applyStateChanges")

    def drawImage(self,*args,**kwds):
        raise NotImplementedError('drawImage')

if __name__=='__main__':
    print("this file has no script interpretation")
    print(__doc__)



#Autogenerated by ReportLab guiedit do not edit
from reportlab.graphics.charts.legends import Legend
from reportlab.graphics.charts.lineplots import ScatterPlot
from reportlab.graphics.shapes import Drawing, _DrawingEditorMixin
from reportlab.graphics.charts.textlabels import Label
from reportlab.graphics.samples.excelcolors import *

class ScatterLines(_DrawingEditorMixin,Drawing):
    def __init__(self,width=200,height=150,*args,**kw):
        Drawing.__init__(self,width,height,*args,**kw)
        self._add(self,ScatterPlot(),name='chart',validate=None,desc="The main chart")
        self.chart.width      = 115
        self.chart.height     = 80
        self.chart.x          = 30
        self.chart.y          = 40
        self.chart.lines[0].strokeColor = color01
        self.chart.lines[1].strokeColor = color02
        self.chart.lines[2].strokeColor = color03
        self.chart.lines[3].strokeColor = color04
        self.chart.lines[4].strokeColor = color05
        self.chart.lines[5].strokeColor = color06
        self.chart.lines[6].strokeColor = color07
        self.chart.lines[7].strokeColor = color08
        self.chart.lines[8].strokeColor = color09
        self.chart.lines[9].strokeColor = color10
        self.chart.lines[0].symbol = None
        self.chart.lines[1].symbol = None
        self.chart.lines[2].symbol = None
        self.chart.lines[3].symbol = None
        self.chart.lines[4].symbol = None
        self.chart.lines[5].symbol = None
        self.chart.lines[6].symbol = None
        self.chart.lines[7].symbol = None
        self.chart.lines[8].symbol = None
        self.chart.lines[9].symbol = None
        self.chart.fillColor         = backgroundGrey
        self.chart.lineLabels.fontName              = 'Helvetica'
        self.chart.xValueAxis.labels.fontName       = 'Helvetica'
        self.chart.xValueAxis.labels.fontSize       = 7
        self.chart.xValueAxis.forceZero             = 0
        self.chart.data             = [((100,100), (200,200), (250,210), (300,300), (400,500)), ((100,200), (200,300), (250,200), (300,400), (400, 600))]
        self.chart.xValueAxis.avoidBoundFrac           = 1
        self.chart.xValueAxis.gridEnd                  = 115
        self.chart.xValueAxis.tickDown                 = 3
        self.chart.xValueAxis.visibleGrid              = 1
        self.chart.yValueAxis.tickLeft              = 3
        self.chart.yValueAxis.labels.fontName       = 'Helvetica'
        self.chart.yValueAxis.labels.fontSize       = 7
        self._add(self,Label(),name='Title',validate=None,desc="The title at the top of the chart")
        self.Title.fontName   = 'Helvetica-Bold'
        self.Title.fontSize   = 7
        self.Title.x          = 100
        self.Title.y          = 135
        self.Title._text      = 'Chart Title'
        self.Title.maxWidth   = 180
        self.Title.height     = 20
        self.Title.textAnchor ='middle'
        self._add(self,Legend(),name='Legend',validate=None,desc="The legend or key for the chart")
        self.Legend.colorNamePairs = [(color01, 'Widgets'), (color02, 'Sprockets')]
        self.Legend.fontName       = 'Helvetica'
        self.Legend.fontSize       = 7
        self.Legend.x              = 153
        self.Legend.y              = 85
        self.Legend.dxTextSpace    = 5
        self.Legend.dy             = 5
        self.Legend.dx             = 5
        self.Legend.deltay         = 5
        self.Legend.alignment      ='right'
        self.chart.lineLabelFormat  = None
        self.chart.xLabel           = 'X Axis'
        self.chart.y                = 30
        self.chart.yLabel           = 'Y Axis'
        self.chart.yValueAxis.gridEnd             = 115
        self.chart.yValueAxis.visibleGrid         = 1
        self.chart.yValueAxis.labelTextFormat     = '%d'
        self.chart.yValueAxis.forceZero           = 1
        self.chart.xValueAxis.forceZero           = 1
        self.chart.joinedLines                    = 1
        self._add(self,0,name='preview',validate=None,desc=None)

if __name__=="__main__": #NORUNTESTS
    ScatterLines().save(formats=['pdf'],outDir=None,fnRoot='scatter_lines')



#Copyright ReportLab Europe Ltd. 2000-2017
#see license.txt for license details
#history https://hg.reportlab.com/hg-public/reportlab/log/tip/src/reportlab/pdfgen/textobject.py
__version__='3.3.0'
__doc__="""
PDFTextObject is an efficient way to add text to a Canvas. Do not
instantiate directly, obtain one from the Canvas instead.

Progress Reports:
8.83, 2000-01-13, gmcm: created from pdfgen.py
"""
from reportlab.lib.colors import Color, CMYKColor, CMYKColorSep, toColor
from reportlab.lib.utils import isBytes, isStr, asUnicode
from reportlab.lib.rl_accel import fp_str
from reportlab.pdfbase.pdfmetrics import getFont as pdfmetrics_getFont, stringWidth as pdfmetrics_stringWidth, unicode2T1 as pdfmetrics_unicode2T1
from reportlab.pdfbase.ttfonts import ShapedStr, ShapeData, _sdGuardL, shapeStr
from itertools import groupby
from operator import itemgetter

#this is to handle the optionality of rlbidi
try:
    import rlbidi
    log2vis = rlbidi.log2vis
    class BidiStr(str):
        '''A str with indices visual __bidiV__, logical __bidiL__'''
        def __new__(cls, s, bidiV=-1, bidiL=-1):
            self = super().__new__(cls,s)
            self.__bidiV__ = bidiV
            self.__bidiL__ = bidiL
            return self
    isBidiStr = lambda _: isinstance(_,BidiStr)
    class BidiList(list):
        '''A list with indices visual __bidiV__, logical __bidiL__'''
        def __init__(self, L, bidiV=-1, bidiL=-1):
            super().__init__(L)
            self.__bidiV__ = bidiV
            self.__bidiL__ = bidiL
        def __repr__(self):
            return f'{self.__class__.__name__}({super().__repr__()},{self.__bidiV__},{self.__bidiL__})'
    isBidiList = lambda _: isinstance(_,BidiList)
    def bidiText(text,direction):
        if direction: direction = direction.upper()
        return log2vis(text, direction,clean=True) if direction in ('LTR','RTL') else text
    from reportlab.lib.utils import KlassStore
    from copy import deepcopy
    _bidiKS = KlassStore()
    def innerBidiStrWrap(s, bidiV=-1, bidiL=-1):
        if isinstance(s,BidiStr): return s
        if type(s) is str:
            klass = BidiStr
        else:
            sklassName = s.__class__.__name__
            klassName = f'BidiIndexed{sklassName}'
            if klassName not in _bidiKS:
                NS=dict(BidiStr=BidiStr,klassName=klassName,klass=s.__class__,deepcopy=deepcopy)
                exec(   f'''class {klassName}(klass,BidiStr):\n'''
                        '''\tdef __new__(cls,s,bidiV=-1,bidiL=-1):\n'''
                        '''\t\tself = super(cls,cls).__new__(cls,s)\n'''
                        '''\t\tif hasattr(s,'__dict__'): self.__dict__=deepcopy(s.__dict__)\n'''
                        '''\t\tself.__bidiV__ = bidiV\n'''
                        '''\t\tself.__bidiL__ = bidiL\n'''
                        '''\t\treturn self\n''',
                        NS)
                _bidiKS.add(klassName,NS[klassName])
            klass = _bidiKS[klassName]
        return klass(s,bidiV=bidiV,bidiL=bidiL)
    def bidiStrWrap(s, orig):
        if not isinstance(orig,BidiStr): return s
        return innerBidiStrWrap(s,orig.__bidiV__,orig.__bidiL__)
    def bidiListWrap(L, orig):
        if not isinstance(orig,(BidiList,BidiIndex)) or isinstance(L,BidiList): return L
        if type(L) is list:
            klass = BidiList
        else:
            lklassName = L.__class__.__name__
            klassName = f'BidiIndexed{lklassName}'
            if klassName not in _bidiKS:
                NS=dict(BidiList=BidiList,klassName=klassName,klass=L.__class__,deepcopy=deepcopy)
                exec(   f'''class {klassName}(klass,BidiList):\n'''
                        '''\tdef __init__(self,L,bidiV=-1,bidiL=-1):\n'''
                        '''\t\tklass.__init__(self,L)\n'''
                        '''\t\tif hasattr(L,'__dict__'): self.__dict__=deepcopy(L.__dict__)\n'''
                        '''\t\tself.__bidiV__ = bidiV\n'''
                        '''\t\tself.__bidiL__ = bidiL\n''',
                        NS)
                _bidiKS.add(klassName,NS[klassName])
            klass = _bidiKS[klassName]
        return klass(L,orig.__bidiV__,orig.__bidiL__)
        #return _bidiKS[klassName](L,orig.__bidiV__,orig.__bidiL__)
    class BidiIndex:
        def __init__(self,bidiV=-1,bidiL=-1):
            self.__bidiV__ = bidiV
            self.__bidiL__ = bidiL
    import re
    wordpat = re.compile(r'[^ ]+',re.M)
    del re
    def bidiWordList(words,direction='RTL', clean=True, wx=False):
        '''takes words (list of strings) returns bidi associated lists
        if wx is True then the V2L index only is returned
        '''
        if direction: direction = direction.upper()
        if direction not in ('LTR','RTL'): return words
        if not isinstance(words,(list,tuple)):
            raise ValueError('bidiWordList argument words should be a list or tuple of strings')
        raw = ' '.join(words)
        V2L = []
        bidi = log2vis(raw, base_direction=direction, clean=clean, positions_V_to_L=V2L)

        VMAP = {}
        for i, m in enumerate(wordpat.finditer(bidi)):
            t = (m.group(0), i)
            start, end = m.span()
            for j in range(start,end):
                VMAP[V2L[j]] = t

        res = [].append
        #create result by assigning a V word to a raw one
        for w, m in enumerate(wordpat.finditer(raw)):
            for j in range(*m.span()):
                if j in VMAP:
                    s, i = VMAP[j]
                    res(i if wx else BidiStr(s,i,w))
                    break
            else:
                #we seem to have a raw word that doesn't appear in bidi
                pass
        return res.__self__
    def bidiShapedText(text, direction='RTL', clean=True, fontName='Helvetica', fontSize=10, shaping=False):
        '''return shaped/bidi text and width; assumes text is aways in logical order'''
        text = asUnicode(text)
        if shaping:
            font = pdfmetrics_getFont(fontName)
            if not font.shapable: shaping = False
        if direction: direction = direction.upper()
        if direction in ('RTL','LTR'):
            if shaping:
                w = text.lstrip()
                bL = len(text) - len(w)
                text = w.rstrip()
                bR = len(w) - len(text)
                LW = [text[slice(*m.span())] for m in wordpat.finditer(text)]   #logical words
                VX = bidiWordList(LW,direction=direction,wx=True)               #visual order
                SW = [shapeStr(LW[i],fontName,fontSize) for i in VX]    #shaped words in visual order
                if len(VX)>1 and VX[0]>VX[-1]:
                    bL, bR = bR, bL
                text = shapeStr(bL*' ',fontName,fontSize, force=True)
                for w in SW:
                    if not hasattr(w,'__shapeData__'): w = shapeStr(w,fontName,fontSize,force=True)
                    if w is not SW[0]: text += shapeStr(' ',fontName,fontSize, force=True)
                    text += w
                if bR:
                    text += shapeStr(bR*' ',fontName,fontSize)
            else:
                text = log2vis(text,base_direction=direction,clean=clean)
        else:
            if shaping:
                text = shapeStr(text, fontName, fontSize)
        width = (sum((_.x_advance for _ in text.__shapeData__))*fontSize/1000 if isinstance(text,ShapedStr)
                    else pdfmetrics_stringWidth(text, fontName, fontSize))
        return text, width
    def bidiFragWord(w,direction=None,bidiV=-1,bidiL=-1, clean=True):
        if direction not in ('RTL','LTR'): return w
        text = ''
        cbd = [].append
        fL = [].append
        for i, (f, s) in enumerate(w[1:]):
            if hasattr(f,'cbDefn'):
                cbd(i)
            else:
                for u in s:
                    fL(i)
                    text += u
        if len(text)<=1: return w #nothing to bidi
        fL = fL.__self__
        V2L = []
        bidi = log2vis(text, base_direction=direction, clean=clean, positions_V_to_L=V2L)
        fsL = [(w[1+k][0],''.join((ks[1] for ks in g))) for k, g in groupby(((fL[V2L[i]],v) for i,v in enumerate(bidi)),key=itemgetter(0))]
        bfw = [sum((pdfmetrics_stringWidth(s,f.fontName,f.fontSize) for f,s in fsL))] + fsL
        cbd = cbd.__self__
        if cbd:
            if V2L[0]>V2L[-1]:
                #reversed
                for i in sorted(cbd):
                    fs = w[i+1]
                    bfw.insert(len(w)-i,fs)
                    bfw[0] += getattr(fs,'width',0)
            else:
                for i in reversed(sorted(cbd)):
                    fs = w[i+1]
                    bfw.insert(i+1,fs)
                    bfw[0] += getattr(fs,'width',0)
        bfw = w.__class__(bfw)
        return bidiListWrap(bfw, BidiIndex(bidiV,bidiL))
    rtlSupport = True
except:
    import warnings
    _rlbidiMsg = 'rlbidi is not installed - RTL/LTR not supported'
    def log2vis(*args,**kwds):
        raise ValueError(_rlbidiMsg)
    isBidiStr = isBidiList = lambda _: False
    BidiStr = str
    BidiList = list
    def bidiText(*args,**kwds):
        direction = kwds.pop('direction',None)
        if direction: direction = direction.upper()
        if direction in ('LTR','RTL'):
            warnings.warn(_rlbidiMsg,stacklevel=0)
        return args[0]
    def bidiShapedText(text, direction='RTL', clean=True, fontName='Helvetica', fontSize=10, shaping=False):
        return bidiText(text,direction), pdfmetrics_stringWidth(text,fontName,fontSize)
    bidiWordList = bidiStrWrap = bidiListWrap = bidiFragWord = innerBidiStrWrap = BidiIndex = bidiText
    rtlSupport = False

class _PDFColorSetter:
    '''Abstracts the color setting operations; used in Canvas and Textobject
    asseumes we have a _code object'''
    def _checkSeparation(self,cmyk):
        if isinstance(cmyk,CMYKColorSep):
            name,sname = self._doc.addColor(cmyk)
            if name not in self._colorsUsed:
                self._colorsUsed[name] = sname
            return name

    #if this is set to a callable(color) --> color it can be used to check color setting
    #see eg _enforceCMYK/_enforceRGB
    _enforceColorSpace = None

    def setFillColorCMYK(self, c, m, y, k, alpha=None):
         """set the fill color useing negative color values
         (cyan, magenta, yellow and darkness value).
         Takes 4 arguments between 0.0 and 1.0"""
         self.setFillColor((c,m,y,k),alpha=alpha)

    def setStrokeColorCMYK(self, c, m, y, k, alpha=None):
         """set the stroke color useing negative color values
            (cyan, magenta, yellow and darkness value).
            Takes 4 arguments between 0.0 and 1.0"""
         self.setStrokeColor((c,m,y,k),alpha=alpha)

    def setFillColorRGB(self, r, g, b, alpha=None):
        """Set the fill color using positive color description
           (Red,Green,Blue).  Takes 3 arguments between 0.0 and 1.0"""
        self.setFillColor((r,g,b),alpha=alpha)

    def setStrokeColorRGB(self, r, g, b, alpha=None):
        """Set the stroke color using positive color description
           (Red,Green,Blue).  Takes 3 arguments between 0.0 and 1.0"""
        self.setStrokeColor((r,g,b),alpha=alpha)

    def setFillColor(self, aColor, alpha=None):
        """Takes a color object, allowing colors to be referred to by name"""
        if self._enforceColorSpace:
            aColor = self._enforceColorSpace(aColor)
        if isinstance(aColor, CMYKColor):
            d = aColor.density
            c,m,y,k = (d*aColor.cyan, d*aColor.magenta, d*aColor.yellow, d*aColor.black)
            self._fillColorObj = aColor
            name = self._checkSeparation(aColor)
            if name:
                self._code.append('/%s cs %s scn' % (name,fp_str(d)))
            else:
                self._code.append('%s k' % fp_str(c, m, y, k))
        elif isinstance(aColor, Color):
            rgb = (aColor.red, aColor.green, aColor.blue)
            self._fillColorObj = aColor
            self._code.append('%s rg' % fp_str(rgb) )
        elif isinstance(aColor,(tuple,list)):
            l = len(aColor)
            if l==3:
                self._fillColorObj = aColor
                self._code.append('%s rg' % fp_str(aColor) )
            elif l==4:
                self._fillColorObj = aColor
                self._code.append('%s k' % fp_str(aColor))
            else:
                raise ValueError('Unknown color %r' % aColor)
        elif isStr(aColor):
            self.setFillColor(toColor(aColor))
        else:
            raise ValueError('Unknown color %r' % aColor)
        if alpha is not None:
            self.setFillAlpha(alpha)
        elif getattr(aColor, 'alpha', None) is not None:
            self.setFillAlpha(aColor.alpha)

    def setStrokeColor(self, aColor, alpha=None):
        """Takes a color object, allowing colors to be referred to by name"""
        if self._enforceColorSpace:
            aColor = self._enforceColorSpace(aColor)
        if isinstance(aColor, CMYKColor):
            d = aColor.density
            c,m,y,k = (d*aColor.cyan, d*aColor.magenta, d*aColor.yellow, d*aColor.black)
            self._strokeColorObj = aColor
            name = self._checkSeparation(aColor)
            if name:
                self._code.append('/%s CS %s SCN' % (name,fp_str(d)))
            else:
                self._code.append('%s K' % fp_str(c, m, y, k))
        elif isinstance(aColor, Color):
            rgb = (aColor.red, aColor.green, aColor.blue)
            self._strokeColorObj = aColor
            self._code.append('%s RG' % fp_str(rgb) )
        elif isinstance(aColor,(tuple,list)):
            l = len(aColor)
            if l==3:
                self._strokeColorObj = aColor
                self._code.append('%s RG' % fp_str(aColor) )
            elif l==4:
                self._strokeColorObj = aColor
                self._code.append('%s K' % fp_str(aColor))
            else:
                raise ValueError('Unknown color %r' % aColor)
        elif isStr(aColor):
            self.setStrokeColor(toColor(aColor))
        else:
            raise ValueError('Unknown color %r' % aColor)
        if alpha is not None:
            self.setStrokeAlpha(alpha)
        elif getattr(aColor, 'alpha', None) is not None:
            self.setStrokeAlpha(aColor.alpha)

    def setFillGray(self, gray, alpha=None):
        """Sets the gray level; 0.0=black, 1.0=white"""
        self._fillColorObj = (gray, gray, gray)
        self._code.append('%s g' % fp_str(gray))
        if alpha is not None:
            self.setFillAlpha(alpha)

    def setStrokeGray(self, gray, alpha=None):
        """Sets the gray level; 0.0=black, 1.0=white"""
        self._strokeColorObj = (gray, gray, gray)
        self._code.append('%s G' % fp_str(gray))
        if alpha is not None:
            self.setFillAlpha(alpha)

    def setStrokeAlpha(self,a):
        if not (isinstance(a,(float,int)) and 0<=a<=1):
            raise ValueError('setStrokeAlpha invalid value %r' % a)
        getattr(self,'_setStrokeAlpha',lambda x: None)(a)

    def setFillAlpha(self,a):
        if not (isinstance(a,(float,int)) and 0<=a<=1):
            raise ValueError('setFillAlpha invalid value %r' % a)
        getattr(self,'_setFillAlpha',lambda x: None)(a)

    def setStrokeOverprint(self,a):
        getattr(self,'_setStrokeOverprint',lambda x: None)(a)

    def setFillOverprint(self,a):
        getattr(self,'_setFillOverprint',lambda x: None)(a)

    def setOverprintMask(self,a):
        getattr(self,'_setOverprintMask',lambda x: None)(a)

class PDFTextObject(_PDFColorSetter):
    """PDF logically separates text and graphics drawing; text
    operations need to be bracketed between BT (Begin text) and
    ET operators. This class ensures text operations are
    properly encapusalted. Ask the canvas for a text object
    with beginText(x, y).  Do not construct one directly.
    Do not use multiple text objects in parallel; PDF is
    not multi-threaded!

    It keeps track of x and y coordinates relative to its origin."""

    def __init__(self, canvas, x=0,y=0, direction=None):
        self._code = ['BT']    #no point in [] then append RGB
        self._canvas = canvas  #canvas sets this so it has access to size info
        self._fontname = self._canvas._fontname
        self._fontsize = self._canvas._fontsize
        self._leading = self._canvas._leading
        self._doc = self._canvas._doc
        self._colorsUsed = self._canvas._colorsUsed
        self._enforceColorSpace = getattr(canvas,'_enforceColorSpace',None)
        font = pdfmetrics_getFont(self._fontname)
        self._curSubset = -1
        self.direction = direction
        self.setTextOrigin(x, y)
        self._textRenderMode = 0
        self._clipping = 0
        self._rise = 0

    def getCode(self):
        "pack onto one line; used internally"
        self._code.append('ET')
        if self._clipping:
            self._code.append('%d Tr' % (self._textRenderMode^4))
        return ' '.join(self._code)

    def setTextOrigin(self, x, y):
        if self._canvas.bottomup:
            self._code.append('1 0 0 1 %s Tm' % fp_str(x, y)) #bottom up
        else:
            self._code.append('1 0 0 -1 %s Tm' % fp_str(x, y))  #top down

        # The current cursor position is at the text origin
        self._x0 = self._x = x
        self._y0 = self._y = y

    def setTextTransform(self, a, b, c, d, e, f):
        "Like setTextOrigin, but does rotation, scaling etc."
        if not self._canvas.bottomup:
            c = -c    #reverse bottom row of the 2D Transform
            d = -d
        self._code.append('%s Tm' % fp_str(a, b, c, d, e, f))

        # The current cursor position is at the text origin Note that
        # we aren't keeping track of all the transform on these
        # coordinates: they are relative to the rotations/sheers
        # defined in the matrix.
        self._x0 = self._x = e
        self._y0 = self._y = f

    def moveCursor(self, dx, dy):
        """Starts a new line at an offset dx,dy from the start of the
        current line. This does not move the cursor relative to the
        current position, and it changes the current offset of every
        future line drawn (i.e. if you next do a textLine() call, it
        will move the cursor to a position one line lower than the
        position specificied in this call.  """

        # Check if we have a previous move cursor call, and combine
        # them if possible.
        if self._code and self._code[-1].endswith(' Td'):
            L = self._code[-1].split()
            if len(L)==3:
                del self._code[-1]
            else:
                self._code[-1] = ''.join(L[:-4])

            # Work out the last movement
            lastDx = float(L[-3])
            lastDy = float(L[-2])

            # Combine the two movement
            dx += lastDx
            dy -= lastDy

            # We will soon add the movement to the line origin, so if
            # we've already done this for lastDx, lastDy, remove it
            # first (so it will be right when added back again).
            self._x0 -= lastDx
            self._y0 -= lastDy

        # Output the move text cursor call.
        self._code.append('%s Td' % fp_str(dx, -dy))

        # Keep track of the new line offsets and the cursor position
        self._x0 += dx
        self._y0 += dy
        self._x = self._x0
        self._y = self._y0

    def setXPos(self, dx):
        """Starts a new line dx away from the start of the
        current line - NOT from the current point! So if
        you call it in mid-sentence, watch out."""
        self.moveCursor(dx,0)

    def getCursor(self):
        """Returns current text position relative to the last origin."""
        return (self._x, self._y)

    def getStartOfLine(self):
        """Returns a tuple giving the text position of the start of the
        current line."""
        return (self._x0, self._y0)

    def getX(self):
        """Returns current x position relative to the last origin."""
        return self._x

    def getY(self):
        """Returns current y position relative to the last origin."""
        return self._y

    def _setFont(self, psfontname, size):
        """Sets the font and fontSize
        Raises a readable exception if an illegal font
        is supplied.  Font names are case-sensitive! Keeps track
        of font anme and size for metrics."""
        self._fontname = psfontname
        self._fontsize = size
        font = pdfmetrics_getFont(self._fontname)

        if font._dynamicFont:
            self._curSubset = -1
        else:
            pdffontname = self._canvas._doc.getInternalFontName(psfontname)
            self._code.append('%s %s Tf' % (pdffontname, fp_str(size)))

    def setFont(self, psfontname, size, leading = None):
        """Sets the font.  If leading not specified, defaults to 1.2 x
        font size. Raises a readable exception if an illegal font
        is supplied.  Font names are case-sensitive! Keeps track
        of font anme and size for metrics."""
        self._fontname = psfontname
        self._fontsize = size
        if leading is None:
            leading = size * 1.2
        self._leading = leading
        font = pdfmetrics_getFont(self._fontname)
        if font._dynamicFont:
            self._curSubset = -1
        else:
            pdffontname = self._canvas._doc.getInternalFontName(psfontname)
            self._code.append('%s %s Tf %s TL' % (pdffontname, fp_str(size), fp_str(leading)))

    def setCharSpace(self, charSpace):
         """Adjusts inter-character spacing"""
         self._charSpace = charSpace
         self._code.append('%s Tc' % fp_str(charSpace))

    def setWordSpace(self, wordSpace):
        """Adjust inter-word spacing.  This can be used
        to flush-justify text - you get the width of the
        words, and add some space between them."""
        self._wordSpace = wordSpace
        self._code.append('%s Tw' % fp_str(wordSpace))

    def setHorizScale(self, horizScale):
        "Stretches text out horizontally"
        self._horizScale = 100 + horizScale
        self._code.append('%s Tz' % fp_str(horizScale))

    def setLeading(self, leading):
        "How far to move down at the end of a line."
        self._leading = leading
        self._code.append('%s TL' % fp_str(leading))

    def setTextRenderMode(self, mode):
        """Set the text rendering mode.

        0 = Fill text
        1 = Stroke text
        2 = Fill then stroke
        3 = Invisible
        4 = Fill text and add to clipping path
        5 = Stroke text and add to clipping path
        6 = Fill then stroke and add to clipping path
        7 = Add to clipping path

        after we start clipping we mustn't change the mode back until after the ET
        """

        assert mode in (0,1,2,3,4,5,6,7), "mode must be in (0,1,2,3,4,5,6,7)"
        if (mode & 4)!=self._clipping:
            mode |= 4
            self._clipping = mode & 4
        if self._textRenderMode!=mode:
            self._textRenderMode = mode
            self._code.append('%d Tr' % mode)

    def setRise(self, rise):
        "Move text baseline up or down to allow superscript/subscripts"
        v = f'{fp_str(rise)} Ts'
        if self._code[-1].endswith(' Ts'): #optimize out r0 Ts r1 Ts 
            #reverse previous changes
            self._y += self._rise
            self._code[-1] = v
        else:
            self._rise = rise
            self._y -= rise
            self._code.append(v)

    def _formatText(self, text):
        "Generates PDF text output operator(s)"
        #if log2vis and self.direction in ('LTR','RTL'):
        #   # Use pyfribidi to write the text in the correct visual order.
        #   text = log2vis(text, self.direction)
        canv = self._canvas
        font = pdfmetrics_getFont(self._fontname)
        state = (self._code, self._x, self._y)
        try:
            self._code = []
            R = self._code.append
            if font._dynamicFont:
                canv_escape = canv._escape
                tmpl = None
                r0 = self._rise
                #it's a truetype font
                if font.shapable and isinstance(text,ShapedStr):
                    sd0 = 0
                    r0 = self._rise
                    shapeData = text.__shapeData__
                    fontsize = self._fontsize
                    for subset, t in font.splitString(text, canv._doc):
                        cluster = None
                        if subset!=self._curSubset:
                            if not tmpl:
                                tmpl = f'{fp_str(fontsize)} Tf {fp_str(self._leading)} TL'
                            R(f'{font.getSubsetInternalName(subset, canv._doc)} {tmpl}')
                            self._curSubset = subset
                        sd1 = sd0 + len(t)
                        SD = shapeData[sd0:sd1] + _sdGuardL
                        sd0 = sd1
                        for i, sd in enumerate(SD):
                            r = r0 + fontsize*sd.y_offset/1000
                            if cluster is None or sd.cluster<0 or r!=self._rise:
                                if cluster is not None:
                                    #end current cluster
                                    A = [v for v in ((('(%s)' % canv_escape(b''.join(g))) if k else fp_str(sum(g)))
                                                for k, g in groupby(filter(None,A.__self__),lambda x: isinstance(x,bytes))) if v!='0']
                                    if len(A)==1 and A[0].startswith('('):
                                        R(f'{A[0]} Tj')
                                    else:
                                        R(f'[{" ".join(A)}] TJ')

                                    if self._rise!=r0: self.setRise(r0)
                                    if sd.cluster<0: break

                                #begin new cluster
                                if r!=self._rise: self.setRise(r)
                                cluster = sd.cluster
                                A = [].append

                            #we assume that both harfbuzz and pdf positions are correct
                            A(-sd.x_offset)     #adjust using harfbuzz offset
                            A(bytes(chr(t[i]).encode('latin1')))
                            #A(sd.x_offset)     #remove the harfbuzz adjustment
                            #we assume the harfbuzz position is correct, but we will have
                            # 1<----O------|
                            # 1-----O ---->|--------------A--------->2
                            #
                            # 1<----O------|<....(W-O).....><...x...>2
                            # 1--------------------W------->
                            # 
                            # W - O + x = A ==> x = A-W+O
                            A(sd.width - sd.x_advance + sd.x_offset)
                    if self._rise!=r0: self.setRise(r0)
                else:
                    for subset, t in font.splitString(text, canv._doc):
                        if subset!=self._curSubset:
                            if not tmpl:
                                tmpl = f'{fp_str(self._fontsize)} Tf {fp_str(self._leading)} TL'
                            R(f'{font.getSubsetInternalName(subset, canv._doc)} {tmpl}')
                            self._curSubset = subset
                        R(f'({canv_escape(t)}) Tj')
            elif font._multiByte:
                #all the fonts should really work like this - let them know more about PDF...
                R("%s %s Tf %s TL" % (
                    canv._doc.getInternalFontName(font.fontName),
                    fp_str(self._fontsize),
                    fp_str(self._leading)
                    ))
                R("(%s) Tj" % font.formatForPdf(text))
            else:
                #convert to T1  coding
                fc = font
                if isBytes(text):
                    try:
                        text = text.decode('utf8')
                    except UnicodeDecodeError as e:
                        i,j = e.args[2:4]
                        raise UnicodeDecodeError(*(e.args[:4]+('%s\n%s-->%s<--%s' % (e.args[4],text[max(i-10,0):i],text[i:j],text[j:j+10]),)))

                canv_escape = canv._escape
                for f, t in pdfmetrics_unicode2T1(text,[font]+font.substitutionFonts):
                    if f!=fc:
                        R("%s %s Tf %s TL" % (canv._doc.getInternalFontName(f.fontName), fp_str(self._fontsize), fp_str(self._leading)))
                        fc = f
                    R(f'({canv_escape(t)}) Tj')
                if font!=fc:
                    R("%s %s Tf %s TL" % (canv._doc.getInternalFontName(self._fontname), fp_str(self._fontsize), fp_str(self._leading)))
        finally:
            self._code, self._x, self._y = state
        return ' '.join(R.__self__)

    def _shapedTextOut(self, text, dx, dy):
        add = self._code.append
        canv = self._canvas
        font = pdfmetrics_getFont(self._fontname)
        canv_escape = canv._escape
        for subset, t in font.splitString(text, canv._doc):
            if subset!=self._curSubset:
                pdffontname = font.getSubsetInternalName(subset, canv._doc)
                R.append("%s %s Tf %s TL" % (pdffontname, fp_str(self._fontsize), fp_str(self._leading)))
                self._curSubset = subset
            if dy:
                print(f'{dy} -->',end='')
                dy = (dy / font.face.unitsPerEm) * self._fontsize
                print(f'{dy}')
                add(f'{fp_str(dy)} Ts')
            if dx:
                add(f'[{fp_str(font.pdfScale(dx))} ({canv_escape(t)})] TJ')
            if dy:
                add(f'{fp_str(-dy)} Ts')
            

    def _textOut(self, text, TStar=0):
        "prints string at current point, ignores text cursor"
        self._code.append('%s%s' % (self._formatText(text), (TStar and ' T*' or '')))

    def textOut(self, text):
        """prints string at current point, text cursor moves across."""
        self._x = self._x + self._canvas.stringWidth(text, self._fontname, self._fontsize)
        self._code.append(self._formatText(text))

    def textLine(self, text=''):
        """prints string at current point, text cursor moves down.
        Can work with no argument to simply move the cursor down."""
        # Update the coordinates of the cursor
        self._x = self._x0
        if self._canvas.bottomup:
            self._y = self._y - self._leading
        else:
            self._y = self._y + self._leading

        # Update the location of the start of the line
        # self._x0 is unchanged
        self._y0 = self._y

        # Output the text followed by a PDF newline command
        self._code.append('%s T*' % self._formatText(text))

    def textLines(self, stuff, trim=1):
        """prints multi-line or newlined strings, moving down.  One
        comon use is to quote a multi-line block in your Python code;
        since this may be indented, by default it trims whitespace
        off each line and from the beginning; set trim=0 to preserve
        whitespace."""
        if isStr(stuff):
            lines = asUnicode(stuff).strip().split(u'\n')
            if trim==1:
                lines = [s.strip() for s in lines]
        elif isinstance(stuff,(tuple,list)):
            lines = stuff
        else:
            assert 1==0, "argument to textlines must be string,, list or tuple"

        # Output each line one at a time. This used to be a long-hand
        # copy of the textLine code, now called as a method.
        for line in lines:
            self.textLine(line)

    def __nonzero__(self):
        'PDFTextObject is true if it has something done after the init'
        return self._code != ['BT']

    def _setFillAlpha(self,v):
        self._canvas._doc.ensureMinPdfVersion('transparency')
        self._canvas._extgstate.set(self,'ca',v)

    def _setStrokeOverprint(self,v):
        self._canvas._extgstate.set(self,'OP',v)

    def _setFillOverprint(self,v):
        self._canvas._extgstate.set(self,'op',v)

    def _setOverprintMask(self,v):
        self._canvas._extgstate.set(self,'OPM',v and 1 or 0)



'''functions for 2D affine transformations'''
__all__ = (
    'nullTransform',
    'translate',
    'scale',
    'rotate',
    'skewX',
    'skewY',
    'mmult',
    'inverse',
    'zTransformPoint',
    'transformPoint',
    'transformPoints',
    'zTransformPoints',
    )
from math import cos, sin, tan, radians

# constructors for matrices:
def nullTransform():
    return (1, 0, 0, 1, 0, 0)

def translate(dx, dy=0):
    return (1, 0, 0, 1, dx, dy)

def scale(sx, sy=1):
    return (sx, 0, 0, sy, 0, 0)

def rotate(angle, cx=0, cy=0):
    a = radians(angle)
    sina = sin(a)
    cosa = cos(a)
    return (cosa, sina, -sina, cosa, cx, cy)

def skewX(angle):
    return (1, 0, tan(radians(angle)), 1, 0, 0)

def skewY(angle):
    return (1, tan(radians(angle)), 0, 1, 0, 0)

def skew(ax, ay=0):
    if ay:
        return mmult(skewX(ax),skewY(ay))
    else:
        return skewX(ax)

def mmult(A, B):
    "A postmultiplied by B"
    # I checked this RGB
    # [a0 a2 a4]    [b0 b2 b4]
    # [a1 a3 a5] *  [b1 b3 b5]
    # [      1 ]    [      1 ]
    #
    return (A[0]*B[0] + A[2]*B[1],
            A[1]*B[0] + A[3]*B[1],
            A[0]*B[2] + A[2]*B[3],
            A[1]*B[2] + A[3]*B[3],
            A[0]*B[4] + A[2]*B[5] + A[4],
            A[1]*B[4] + A[3]*B[5] + A[5])

def inverse(A):
    "For A affine 2D represented as 6vec return 6vec version of A**(-1)"
    # I checked this RGB
    det = float(A[0]*A[3] - A[2]*A[1])
    R = [A[3]/det, -A[1]/det, -A[2]/det, A[0]/det]
    return tuple(R+[-R[0]*A[4]-R[2]*A[5],-R[1]*A[4]-R[3]*A[5]])

def zTransformPoint(A,v):
    "Apply the homogenous part of atransformation a to vector v --> A*v"
    return (A[0]*v[0]+A[2]*v[1],A[1]*v[0]+A[3]*v[1])

def transformPoint(A,v):
    "Apply transformation a to vector v --> A*v"
    return (A[0]*v[0]+A[2]*v[1]+A[4],A[1]*v[0]+A[3]*v[1]+A[5])

def transformPoints(matrix, V):
    r = [transformPoint(matrix,v) for v in V]
    if isinstance(V,tuple): r = tuple(r)
    return r

def zTransformPoints(matrix, V):
    return list(map(lambda x,matrix=matrix: zTransformPoint(matrix,x), V))



#!/bin/env python
#Copyright ReportLab Europe Ltd. 2000-2017
#see license.txt for license details
#history https://hg.reportlab.com/hg-public/reportlab/log/tip/src/reportlab/lib/randomtext.py

__version__='3.3.0'

###############################################################################
#   generates so-called 'Greek Text' for use in filling documents.
###############################################################################
__doc__="""Like Lorem Ipsum, but more fun and extensible.

This module exposes a function randomText() which generates paragraphs.
These can be used when testing out document templates and stylesheets.
A number of 'themes' are provided - please contribute more!
We need some real Greek text too.

There are currently six themes provided:
    STARTUP (words suitable for a business plan - or not as the case may be),
    COMPUTERS (names of programming languages and operating systems etc),
    BLAH (variations on the word 'blah'),
    BUZZWORD (buzzword bingo),
    STARTREK (Star Trek),
    PRINTING (print-related terms)
    PYTHON (snippets and quotes from Monty Python)
    CHOMSKY (random lingusitic nonsense)

EXAMPLE USAGE:
    from reportlab.lib import randomtext
    print randomtext.randomText(randomtext.PYTHON, 10)

    This prints a random number of random sentences (up to a limit
    of ten) using the theme 'PYTHON'.

"""

#theme one :-)
STARTUP = ['strategic', 'direction', 'proactive', 'venture capital',
    'reengineering', 'forecast', 'resources', 'SWOT analysis',
    'forward-thinking', 'profit', 'growth', 'doubletalk', 'B2B', 'B2C',
    'venture capital', 'IPO', "NASDAQ meltdown - we're all doomed!"]

#theme two - computery things.
COMPUTERS = ['Python', 'Perl', 'Pascal', 'Java', 'Javascript',
    'VB', 'Basic', 'LISP', 'Fortran', 'ADA', 'APL', 'C', 'C++',
    'assembler', 'Larry Wall', 'Guido van Rossum', 'XML', 'HTML',
    'cgi', 'cgi-bin', 'Amiga', 'Macintosh', 'Dell', 'Microsoft',
    'firewall', 'server', 'Linux', 'Unix', 'MacOS', 'BeOS', 'AS/400',
    'sendmail', 'TCP/IP', 'SMTP', 'RFC822-compliant', 'dynamic',
    'Internet', 'A/UX', 'Amiga OS', 'BIOS', 'boot managers', 'CP/M',
    'DOS', 'file system', 'FreeBSD', 'Freeware', 'GEOS', 'GNU',
    'Hurd', 'Linux', 'Mach', 'Macintosh OS', 'mailing lists', 'Minix',
    'Multics', 'NetWare', 'NextStep', 'OS/2', 'Plan 9', 'Realtime',
    'UNIX', 'VMS', 'Windows', 'X Windows', 'Xinu', 'security', 'Intel',
    'encryption', 'PGP' , 'software', 'ActiveX', 'AppleScript', 'awk',
    'BETA', 'COBOL', 'Delphi', 'Dylan', 'Eiffel', 'extreme programming',
    'Forth', 'Fortran', 'functional languages', 'Guile', 'format your hard drive',
    'Icon', 'IDL', 'Infer', 'Intercal', 'J', 'Java', 'JavaScript', 'CD-ROM',
    'JCL', 'Lisp', '"literate programming"', 'Logo', 'MUMPS', 'C: drive',
    'Modula-2', 'Modula-3', 'Oberon', 'Occam', 'OpenGL', 'parallel languages',
    'Pascal', 'Perl', 'PL/I', 'PostScript', 'Prolog', 'hardware', 'Blue Screen of Death',
    'Rexx', 'RPG', 'Scheme', 'scripting languages', 'Smalltalk', 'crash!', 'disc crash',
    'Spanner', 'SQL', 'Tcl/Tk', 'TeX', 'TOM', 'Visual', 'Visual Basic', '4GL',
    'VRML', 'Virtual Reality Modeling Language', 'difference engine', '...went into "yo-yo mode"',
    'Sun', 'Sun Microsystems', 'Hewlett Packard', 'output device',
    'CPU', 'memory', 'registers', 'monitor', 'TFT display', 'plasma screen',
    'bug report', '"mis-feature"', '...millions of bugs!', 'pizza',
    '"illiterate programming"','...lots of pizza!', 'pepperoni pizza',
    'coffee', 'Jolt Cola[TM]', 'beer', 'BEER!']

#theme three - 'blah' - for when you want to be subtle. :-)
BLAH = ['Blah', 'BLAH', 'blahblah', 'blahblahblah', 'blah-blah',
    'blah!', '"Blah Blah Blah"', 'blah-de-blah', 'blah?', 'blah!!!',
    'blah...', 'Blah.', 'blah;', 'blah, Blah, BLAH!', 'Blah!!!']

#theme four - 'buzzword bingo' time!
BUZZWORD = ['intellectual capital', 'market segment', 'flattening',
        'regroup', 'platform', 'client-based', 'long-term', 'proactive',
        'quality vector', 'out of the loop', 'implement',
        'streamline', 'cost-centered', 'phase', 'synergy',
        'synergize', 'interactive', 'facilitate',
        'appropriate', 'goal-setting', 'empowering', 'low-risk high-yield',
        'peel the onion', 'goal', 'downsize', 'result-driven',
        'conceptualize', 'multidisciplinary', 'gap analysis', 'dysfunctional',
        'networking', 'knowledge management', 'goal-setting',
        'mastery learning', 'communication', 'real-estate', 'quarterly',
        'scalable', 'Total Quality Management', 'best of breed',
        'nimble', 'monetize', 'benchmark', 'hardball',
        'client-centered', 'vision statement', 'empowerment',
        'lean & mean', 'credibility', 'synergistic',
        'backward-compatible', 'hardball', 'stretch the envelope',
        'bleeding edge', 'networking', 'motivation', 'best practice',
        'best of breed', 'implementation', 'Total Quality Management',
        'undefined', 'disintermediate', 'mindset', 'architect',
        'gap analysis', 'morale', 'objective', 'projection',
        'contribution', 'proactive', 'go the extra mile', 'dynamic',
        'world class', 'real estate', 'quality vector', 'credibility',
        'appropriate', 'platform', 'projection', 'mastery learning',
        'recognition', 'quality', 'scenario', 'performance based',
        'solutioning', 'go the extra mile', 'downsize', 'phase',
        'networking', 'experiencing slippage', 'knowledge management',
        'high priority', 'process', 'ethical', 'value-added', 'implement',
        're-factoring', 're-branding', 'embracing change']

#theme five - Star Trek
STARTREK = ['Starfleet', 'Klingon', 'Romulan', 'Cardassian', 'Vulcan',
    'Benzite', 'IKV Pagh', 'emergency transponder', 'United Federation of Planets',
    'Bolian', "K'Vort Class Bird-of-Prey", 'USS Enterprise', 'USS Intrepid',
    'USS Reliant', 'USS Voyager', 'Starfleet Academy', 'Captain Picard',
    'Captain Janeway', 'Tom Paris', 'Harry Kim', 'Counsellor Troi',
    'Lieutenant Worf', 'Lieutenant Commander Data', 'Dr. Beverly Crusher',
    'Admiral Nakamura', 'Irumodic Syndrome', 'Devron system', 'Admiral Pressman',
    'asteroid field', 'sensor readings', 'Binars', 'distress signal', 'shuttlecraft',
    'cloaking device', 'shuttle bay 2', 'Dr. Pulaski', 'Lwaxana Troi', 'Pacifica',
    'William Riker', "Chief O'Brian", 'Soyuz class science vessel', 'Wolf-359',
    'Galaxy class vessel', 'Utopia Planitia yards', 'photon torpedo', 'Archer IV',
    'quantum flux', 'spacedock', 'Risa', 'Deep Space Nine', 'blood wine',
    'quantum torpedoes', 'holodeck', 'Romulan Warbird', 'Betazoid', 'turbolift', 'battle bridge',
    'Memory Alpha', '...with a phaser!', 'Romulan ale', 'Ferrengi', 'Klingon opera',
    'Quark', 'wormhole', 'Bajoran', 'cruiser', 'warship', 'battlecruiser', '"Intruder alert!"',
    'scout ship', 'science vessel', '"Borg Invasion imminent!" ', '"Abandon ship!"',
    'Red Alert!', 'warp-core breech', '"All hands abandon ship! This is not a drill!"']

#theme six - print-related terms
PRINTING = ['points', 'picas', 'leading', 'kerning', 'CMYK', 'offset litho',
    'type', 'font family', 'typography', 'type designer',
    'baseline', 'white-out type', 'WOB', 'bicameral', 'bitmap',
    'blockletter', 'bleed', 'margin', 'body', 'widow', 'orphan',
    'cicero', 'cursive', 'letterform', 'sidehead', 'dingbat', 'leader',
    'DPI', 'drop-cap', 'paragraph', 'En', 'Em', 'flush left', 'left justified',
    'right justified', 'centered', 'italic', 'Latin letterform', 'ligature',
    'uppercase', 'lowercase', 'serif', 'sans-serif', 'weight', 'type foundry',
    'fleuron', 'folio', 'gutter', 'whitespace', 'humanist letterform', 'caption',
    'page', 'frame', 'ragged setting', 'flush-right', 'rule', 'drop shadows',
    'prepress', 'spot-colour', 'duotones', 'colour separations', 'four-colour printing',
    'Pantone[TM]', 'service bureau', 'imagesetter']

#it had to be done!...
#theme seven - the "full Monty"!
PYTHON = ['Good evening ladies and Bruces','I want to buy some cheese', 'You do have some cheese, do you?',
          "Of course sir, it's a cheese shop sir, we've got...",'discipline?... naked? ... With a melon!?',
          'The Church Police!!' , "There's a dead bishop on the landing", 'Would you like a twist of lemming sir?',
          '"Conquistador Coffee brings a new meaning to the word vomit"','Your lupins please',
          'Crelm Toothpaste, with the miracle ingredient Fraudulin',
          "Well there's the first result and the Silly Party has held Leicester.",
          'Hello, I would like to buy a fish license please', "Look, it's people like you what cause unrest!",
          "When we got home, our Dad would thrash us to sleep with his belt!", 'Luxury', "Gumby Brain Specialist",
          "My brain hurts!!!", "My brain hurts too.", "How not to be seen",
          "In this picture there are 47 people. None of them can be seen",
          "Mrs Smegma, will you stand up please?",
          "Mr. Nesbitt has learned the first lesson of 'Not Being Seen', not to stand up.",
          "My hovercraft is full of eels", "Ah. You have beautiful thighs.", "My nipples explode with delight",
          "Drop your panties Sir William, I cannot wait 'til lunchtime",
          "I'm a completely self-taught idiot.", "I always wanted to be a lumberjack!!!",
          "Told you so!! Oh, coitus!!", "",
          "Nudge nudge?", "Know what I mean!", "Nudge nudge, nudge nudge?", "Say no more!!",
          "Hello, well it's just after 8 o'clock, and time for the penguin on top of your television set to explode",
          "Oh, intercourse the penguin!!", "Funny that penguin being there, isn't it?",
          "I wish to register a complaint.", "Now that's what I call a dead parrot", "Pining for the fjords???",
          "No, that's not dead, it's ,uhhhh, resting", "This is an ex-parrot!!",
          "That parrot is definitely deceased.", "No, no, no - it's spelt Raymond Luxury Yach-t, but it's pronounced 'Throatwobbler Mangrove'.",
          "You're a very silly man and I'm not going to interview you.", "No Mungo... never kill a customer."
          "And I'd like to conclude by putting my finger up my nose",
          "egg and Spam", "egg bacon and Spam", "egg bacon sausage and Spam", "Spam bacon sausage and Spam",
          "Spam egg Spam Spam bacon and Spam", "Spam sausage Spam Spam Spam bacon Spam tomato and Spam",
          "Spam Spam Spam egg and Spam", "Spam Spam Spam Spam Spam Spam baked beans Spam Spam Spam",
          "Spam!!", "I don't like Spam!!!", "You can't have egg, bacon, Spam and sausage without the Spam!",
          "I'll have your Spam. I Love it!",
          "I'm having Spam Spam Spam Spam Spam Spam Spam baked beans Spam Spam Spam and Spam",
          "Have you got anything without Spam?", "There's Spam egg sausage and Spam, that's not got much Spam in it.",
          "No one expects the Spanish Inquisition!!", "Our weapon is surprise, surprise and fear!",
          "Get the comfy chair!", "Amongst our weaponry are such diverse elements as: fear, surprise, ruthless efficiency, an almost fanatical devotion to the Pope, and nice red uniforms - Oh damn!",
          "Nobody expects the... Oh bugger!", "What swims in the sea and gets caught in nets? Henri Bergson?",
          "Goats. Underwater goats with snorkels and flippers?", "A buffalo with an aqualung?",
          "Dinsdale was a looney, but he was a happy looney.", "Dinsdale!!",
          "The 127th Upper-Class Twit of the Year Show", "What a great Twit!",
          "thought by many to be this year's outstanding twit",
          "...and there's a big crowd here today to see these prize idiots in action.",
          "And now for something completely different.", "Stop that, it's silly",
          "We interrupt this program to annoy you and make things generally irritating",
          "This depraved and degrading spectacle is going to stop right now, do you hear me?",
          "Stop right there!", "This is absolutely disgusting and I'm not going to stand for it",
          "I object to all this sex on the television. I mean, I keep falling off",
          "Right! Stop that, it's silly. Very silly indeed", "Very silly indeed", "Lemon curry?",
          "And now for something completely different, a man with 3 buttocks",
          "I've heard of unisex, but I've never had it", "That's the end, stop the program! Stop it!"]
leadins=[
    "To characterize a linguistic level L,",
    "On the other hand,",
    "This suggests that",
    "It appears that",
    "Furthermore,",
    "We will bring evidence in favor of the following thesis: ",
    "To provide a constituent structure for T(Z,K),",
    "From C1, it follows that",
    "For any transformation which is sufficiently diversified in application to be of any interest,",
    "Analogously,",
    "Clearly,",
    "Note that",
    "Of course,",
    "Suppose, for instance, that",
    "Thus",
    "With this clarification,",
    "Conversely,",
    "We have already seen that",
    "By combining adjunctions and certain deformations,",
    "I suggested that these results would follow from the assumption that",
    "If the position of the trace in (99c) were only relatively inaccessible to movement,",
    "However, this assumption is not correct, since",
    "Comparing these examples with their parasitic gap counterparts in (96) and (97), we see that",
    "In the discussion of resumptive pronouns following (81),",
    "So far,",
    "Nevertheless,",
    "For one thing,",
    "Summarizing, then, we assume that",
    "A consequence of the approach just outlined is that",
    "Presumably,",
    "On our assumptions,",
    "It may be, then, that",
    "It must be emphasized, once again, that",
    "Let us continue to suppose that",
    "Notice, incidentally, that",
    "A majority  of informed linguistic specialists agree that",
    "There is also a different approach to the [unification] problem,",
    "This approach divorces the cognitive sciences from a biological setting,",
    "The approach relies on the \"Turing Test,\" devised by mathematician Alan Turing,",
    "Adopting this approach,",
    "There is no fact, no meaningful question to be answered,",
    "Another superficial similarity is the interest in simulation of behavior,",
    "A lot of sophistication has been developed about the utilization of machines for complex purposes,",
    ]
 
subjects = [
    "the notion of level of grammaticalness",
    "a case of semigrammaticalness of a different sort",
    "most of the methodological work in modern linguistics",
    "a subset of English sentences interesting on quite independent grounds",
    "the natural general principle that will subsume this case",
    "an important property of these three types of EC",
    "any associated supporting element",
    "the appearance of parasitic gaps in domains relatively inaccessible to ordinary extraction",
    "the speaker-hearer's linguistic intuition",
    "the descriptive power of the base component",
    "the earlier discussion of deviance",
    "this analysis of a formative as a pair of sets of features",
    "this selectionally introduced contextual feature",
    "a descriptively adequate grammar",
    "the fundamental error of regarding functional notions as categorial",
    "relational information",
    "the systematic use of complex symbols",
    "the theory of syntactic features developed earlier",
    ]
 
verbs= [
    "can be defined in such a way as to impose",
    "delimits",
    "suffices to account for",
    "cannot be arbitrary in",
    "is not subject to",
    "does not readily tolerate",
    "raises serious doubts about",
    "is not quite equivalent to",
    "does not affect the structure of",
    "may remedy and, at the same time, eliminate",
    "is not to be considered in determining",
    "is to be regarded as",
    "is unspecified with respect to",
    "is, apparently, determined by",
    "is necessary to impose an interpretation on",
    "appears to correlate rather closely with",
    "is rather different from",
    ]

objects = [
    "problems of phonemic and morphological analysis.",
    "a corpus of utterance tokens upon which conformity has been defined by the paired utterance test.",
    "the traditional practice of grammarians.",
    "the levels of acceptability from fairly high (e.g. (99a)) to virtual gibberish (e.g. (98d)).",
    "a stipulation to place the constructions into these various categories.",
    "a descriptive fact.",
    "a parasitic gap construction.",
    "the extended c-command discussed in connection with (34).",
    "the ultimate standard that determines the accuracy of any proposed grammar.",
    "the system of base rules exclusive of the lexicon.",
    "irrelevant intervening contexts in selectional rules.",
    "nondistinctness in the sense of distinctive feature theory.",
    "a general convention regarding the forms of the grammar.",
    "an abstract underlying order.",
    "an important distinction in language use.",
    "the requirement that branching is not tolerated within the dominance scope of a complex symbol.",
    "the strong generative capacity of the theory.",
    ]

def format_wisdom(text,line_length=72):
    try:
        import textwrap
        return textwrap.fill(text, line_length)
    except:
        return text

def chomsky(times = 1):
    if not isinstance(times, int):
        return format_wisdom(__doc__)
    import random
    prevparts = []
    newparts = []
    output = []
    for i in range(times):
        for partlist in (leadins, subjects, verbs, objects):
            while 1:
                part = random.choice(partlist)
                if part not in prevparts:
                    break
            newparts.append(part)
        output.append(' '.join(newparts))
        prevparts = newparts
        newparts = []
    return format_wisdom('  '.join(output))

def randomText(theme=STARTUP, sentences=5):
    #this may or may not be appropriate in your company
    if type(theme)==type(''):
        if theme.lower()=='chomsky': return chomsky(sentences)
        elif theme.upper() in ('STARTUP','COMPUTERS','BLAH','BUZZWORD','STARTREK','PRINTING','PYTHON'):
            theme = globals()[theme.upper()]
        else:
            raise ValueError('Unknown theme "%s"' % theme)

    from random import randint, choice

    RANDOMWORDS = theme

    #sentences = 5
    output = ""
    for sentenceno in range(randint(1,sentences)):
        output = output + 'Blah'
        for wordno in range(randint(10,25)):
            if randint(0,4)==0:
                word = choice(RANDOMWORDS)
            else:
                word = 'blah'
            output = output + ' ' +word
        output = output+'. '
    return output

if __name__=='__main__':
    import sys, random
    from reportlab.rl_config import invariant as rl_invariant
    if rl_invariant:
        random.seed(1854640162)
        print(f'{" ".join((str(random.randrange(100)) for _ in range(10)))}')
    argv = sys.argv[1:]
    if argv:
        theme = argv.pop(0)
        if argv:
            sentences = int(argv.pop(0))
        else:
            sentences = 5
        try:
            print(randomText(theme,sentences))
        except:
            sys.stderr.write("Usage: randomtext.py [theme [#sentences]]\n")
            sys.stderr.write(" theme in chomsky|STARTUP|COMPUTERS|BLAH|BUZZWORD|STARTREK|PRINTING|PYTHON\n")
            raise
    else:
        print(chomsky(5))



#Copyright ReportLab Europe Ltd. 2000-2017
#see license.txt for license details
'''default settings for reportlab

to override these drop a module rl_local_settings.py parallel to this file or
anywhere on the path.
'''
import os, sys
__version__='3.3.0'
__all__=tuple('''allowTableBoundsErrors
shapeChecking
defaultEncoding
defaultGraphicsFontName
pageCompression
useA85
defaultPageSize
defaultImageCaching
warnOnMissingFontGlyphs
verbose
showBoundary
emptyTableAction
invariant
eps_preview_transparent
eps_preview
eps_ttf_embed
eps_ttf_embed_uid
overlapAttachedSpace
longTableOptimize
autoConvertEncoding
_FUZZ
wrapA85
fsEncodings
odbc_driver
platypus_link_underline
canvas_basefontname
allowShortTableRows
imageReaderFlags
paraFontSizeHeightOffset
canvas_baseColor
ignoreContainerActions
ttfAsciiReadable
pdfMultiLine
pdfComments
debug
listWrapOnFakeWidth
T1SearchPath
TTFSearchPath
CMapSearchPath
decimalSymbol
errorOnDuplicatePageLabelPage
autoGenerateMissingTTFName
allowTTFSubsetting
spaceShrinkage
underlineWidth
underlineOffset
underlineGap
strikeWidth
strikeOffset
strikeGap
hyphenationLang
uriWasteReduce
embeddedHyphenation
hyphenationMinWordLength
reserveTTFNotdef
documentLang
encryptionStrength
trustedHosts
trustedSchemes
renderPMBackend
xmlParser
textPaths
toColorCanUse
defCWRF
unShapedFontGlob'''.split())

allowTableBoundsErrors =    1 # set to 0 to die on too large elements in tables in debug (recommend 1 for production use)
shapeChecking =             1
defaultEncoding =           'WinAnsiEncoding'       # 'WinAnsi' or 'MacRoman'
defaultGraphicsFontName=    'Times-Roman'           #initializer for STATE_DEFAULTS in shapes.py
pageCompression =           1                       # default page compression mode
useA85 =                    1                       #set to 0 to disable Ascii Base 85 stream filters
defaultPageSize =           'A4'                    #default page size
defaultImageCaching =       0                       #set to zero to remove those annoying cached images
warnOnMissingFontGlyphs =   0                       #if 1, warns of each missing glyph
verbose =                   0
showBoundary =              0                       # turns on and off boundary behaviour in Drawing
emptyTableAction=           'error'                 # one of 'error', 'indicate', 'ignore'
invariant=                  0                       #produces repeatable,identical PDFs with same timestamp info (for regression testing)
eps_preview_transparent=    None                    #set to white etc
eps_preview=                1                       #set to False to disable
eps_ttf_embed=              1                       #set to False to disable
eps_ttf_embed_uid=          0                       #set to 1 to enable
overlapAttachedSpace=       1                       #if set non false then adajacent flowable space after
                                                    #and space before are merged (max space is used).
longTableOptimize =         1                       #default do use Henning von Bargen's long table optimizations
autoConvertEncoding  =      0                       #convert internally as needed (experimental)
_FUZZ=                      1e-6                    #fuzz for layout arithmetic
wrapA85=                    0                       #set to 1 to get old wrapped line behaviour
fsEncodings=('utf8','cp1252','cp430')               #encodings to attempt utf8 conversion with
odbc_driver=                'odbc'                  #default odbc driver
platypus_link_underline=    0                       #paragraph links etc underlined if true
canvas_basefontname=        'Helvetica'             #this is used to initialize the canvas; if you override to make
                                                    #something else you are responsible for ensuring the font is registered etc etc
                                                    #this will be used everywhere and the font family connections will be made
                                                    #if the bold/italic/bold italic fonts are also registered and defined as a family.

allowShortTableRows=1                               #allows some rows in a table to be short
imageReaderFlags=0                                  #no longer in use
paraFontSizeHeightOffset=   1                       #if true paragraphs start at height-fontSize
canvas_baseColor=           None                    #initialize the canvas fill and stroke colors if this is set
ignoreContainerActions=     1                       #if true then action flowables in flowable _Containers will be ignored
ttfAsciiReadable=           1                       #smaller subsets when set to 0
pdfMultiLine=               0                       #use more lines in pdf etc
pdfComments=                0                       #put in pdf comments
debug=                      0                       #for debugging code
listWrapOnFakeWidth=        1                       #set to 0/False to force platypus.flowables._listWrapOn to report correct widths
                                                    #else it reports minimum(required,available) width

underlineWidth=             ''                      #empty to use canvas strokeWidth or a distance or number*<letter>
                                                    #   num * <letter> make value proportional to a font size
                                                    #   P paragraph font size
                                                    #   L line max font size
                                                    #   f first use font size
                                                    #   F max fontsize in the tag

underlineOffset=            '-0.125*F'              #fraction of fontSize from baseline to draw underlines at.
underlineGap=               '1'                     #gap for double/triple underline

strikeWidth=                ''
strikeOffset=               '0.25*F'                #fraction of fontSize from baseline to draw strike through at.
strikeGap=                  '1'                     #gap for double/triple strike

                                                    #by default typical value 0.05. may be overridden on a parastyle.
decimalSymbol=              '.'                     #what we use to align floats numerically
errorOnDuplicatePageLabelPage= 0                    #if True will cause repeated PageLabel page numbers to raise an error.
autoGenerateMissingTTFName=0                        #if true we try to auto generate any missing TTF font name

allowTTFSubsetting=         []                      #list of font file names that will be subsetted even when they
                                                    #have the no subsetting flag set. These should be fonts for which
                                                    #the user has explicit permission from the rights holder(s). 
                                                    #This flag could already be overcome by hacking the code.
                                                    #ReportLab takes no responsibility for the use of this setting.

spaceShrinkage=0.05                                 #allowable space shrinkage to make lines fit
hyphenationLang=''                                  #if pyphen installed set this to the language of your choice
                                                    #eg 'en_GB'

uriWasteReduce=0                                    #split URI if we would waste 0.3 of a line or if the URI#
                                                    #would not fit on the next line; if zero then no splitting
                                                    #is attempted. suggested value = 0.3
embeddedHyphenation=0                               #if true attempt hypenation of words with embedded hyphens
hyphenationMinWordLength=5                          #minimum length of words that can be hyphenated
reserveTTFNotdef=1                                  #if true force subset element 0 to be zero(.notdef)
                                                    #helps to fix bug in edge; this is now ignored in code
                                                    #PDFUA forbids index 0(.notdef) in strings
documentLang=None                                   #pdf document catalog Lang value xx-xx not ee_xx
encryptionStrength=40                               #the bits for standard encryption 40, 128 or 256 (AES)
trustedHosts=None                                   #set to a list of trusted for access hosts None means
                                                    #all are trusted glob patterns eg *.reportlab.com are
                                                    #allowed. In environment use a comma separated string.
trustedSchemes=['file', 'rml', 'data', 'https',     #these url schemes are trusted
                'http', 'ftp']
renderPMBackend='rlPyCairo'                         #or '_renderPM' if available
xmlParser='lxml'                                    #or 'pyrxp' for preferred xml parsing
textPaths='freetype'                                #freetype or _renderPM or backend
                                                    #determines what code is used to create Paths from str
                                                    #see reportlab/graphics/utils.py for full horror
toColorCanUse='rl_extended_literal_eval'            #change to None or 'rl_safe_eval' depending on trust
defCWRF=0.02                                        #fraction we can reduce defined column widths for overcommitted
                                                    #undefined widths
unShapedFontGlob=None                               #None or space list of glob patterns that force off shaping

# places to look for T1Font information
T1SearchPath =  (
                'c:/Program Files/Adobe/Acrobat 9.0/Resource/Font', 
                'c:/Program Files/Adobe/Acrobat 8.0/Resource/Font', 
                'c:/Program Files/Adobe/Acrobat 7.0/Resource/Font', 
                'c:/Program Files/Adobe/Acrobat 6.0/Resource/Font', #Win32, Acrobat 6
                'c:/Program Files/Adobe/Acrobat 5.0/Resource/Font', #Win32, Acrobat 5
                'c:/Program Files/Adobe/Acrobat 4.0/Resource/Font', #Win32, Acrobat 4
                '%(disk)s/Applications/Python %(sys_version)s/reportlab/fonts', #Mac?
                '/usr/lib/Acrobat9/Resource/Font',      #Linux, Acrobat 5?
                '/usr/lib/Acrobat8/Resource/Font',      #Linux, Acrobat 5?
                '/usr/lib/Acrobat7/Resource/Font',      #Linux, Acrobat 5?
                '/usr/lib/Acrobat6/Resource/Font',      #Linux, Acrobat 5?
                '/usr/lib/Acrobat5/Resource/Font',      #Linux, Acrobat 5?
                '/usr/lib/Acrobat4/Resource/Font',      #Linux, Acrobat 4
                '/usr/local/Acrobat9/Resource/Font',    #Linux, Acrobat 5?
                '/usr/local/Acrobat8/Resource/Font',    #Linux, Acrobat 5?
                '/usr/local/Acrobat7/Resource/Font',    #Linux, Acrobat 5?
                '/usr/local/Acrobat6/Resource/Font',    #Linux, Acrobat 5?
                '/usr/local/Acrobat5/Resource/Font',    #Linux, Acrobat 5?
                '/usr/local/Acrobat4/Resource/Font',    #Linux, Acrobat 4
                '/usr/share/fonts/default/Type1',       #Linux, Fedora
                '%(REPORTLAB_DIR)s/fonts',              #special
                '%(REPORTLAB_DIR)s/../fonts',           #special
                '%(REPORTLAB_DIR)s/../../fonts',        #special
                '%(CWD)s/fonts',                        #special
                '~/fonts',
                '~/.fonts',
                '%(XDG_DATA_HOME)s/fonts',
                '~/.local/share/fonts',
                 )

# places to look for TT Font information
TTFSearchPath = (
                'c:/winnt/fonts',
                'c:/windows/fonts',
                '/usr/lib/X11/fonts/TrueType/',
                '/usr/share/fonts/truetype',
                '/usr/share/fonts',             #Linux, Fedora
                '/usr/share/fonts/dejavu',      #Linux, Fedora
                '%(REPORTLAB_DIR)s/fonts',      #special
                '%(REPORTLAB_DIR)s/../fonts',   #special
                '%(REPORTLAB_DIR)s/../../fonts',#special
                '%(CWD)s/fonts',                #special
                '~/fonts',
                '~/.fonts',
                '%(XDG_DATA_HOME)s/fonts',
                '~/.local/share/fonts',
                #mac os X - from
                #http://developer.apple.com/technotes/tn/tn2024.html
                '~/Library/Fonts',
                '/Library/Fonts',
                '/Network/Library/Fonts',
                '/System/Library/Fonts',
                )

# places to look for CMap files - should ideally merge with above
CMapSearchPath = (
                  '/usr/lib/Acrobat9/Resource/CMap',
                  '/usr/lib/Acrobat8/Resource/CMap',
                  '/usr/lib/Acrobat7/Resource/CMap',
                  '/usr/lib/Acrobat6/Resource/CMap',
                  '/usr/lib/Acrobat5/Resource/CMap',
                  '/usr/lib/Acrobat4/Resource/CMap',
                  '/usr/local/Acrobat9/Resource/CMap',
                  '/usr/local/Acrobat8/Resource/CMap',
                  '/usr/local/Acrobat7/Resource/CMap',
                  '/usr/local/Acrobat6/Resource/CMap',
                  '/usr/local/Acrobat5/Resource/CMap',
                  '/usr/local/Acrobat4/Resource/CMap',
                  'C:\\Program Files\\Adobe\\Acrobat\\Resource\\CMap',
                  'C:\\Program Files\\Adobe\\Acrobat 9.0\\Resource\\CMap',
                  'C:\\Program Files\\Adobe\\Acrobat 8.0\\Resource\\CMap',
                  'C:\\Program Files\\Adobe\\Acrobat 7.0\\Resource\\CMap',
                  'C:\\Program Files\\Adobe\\Acrobat 6.0\\Resource\\CMap',
                  'C:\\Program Files\\Adobe\\Acrobat 5.0\\Resource\\CMap',
                  'C:\\Program Files\\Adobe\\Acrobat 4.0\\Resource\\CMap',
                  '%(REPORTLAB_DIR)s/fonts/CMap',       #special
                  '%(REPORTLAB_DIR)s/../fonts/CMap',    #special
                  '%(REPORTLAB_DIR)s/../../fonts/CMap', #special
                  '%(CWD)s/fonts/CMap',             #special
                  '%(CWD)s/fonts',              #special
                  '~/fonts/CMap',
                  '~/.fonts/CMap',
                  '%(XDG_DATA_HOME)s/fonts/CMap',
                  '~/.local/share/fonts/CMap',
                  )

if sys.platform.startswith('linux'):
    def _findFontDirs(*ROOTS):
        R = [].append
        for rootd in ROOTS:
            for root, dirs, files in os.walk(rootd):
                if not files: continue
                R(root)
        return tuple(R.__self__)
    T1SearchPath = T1SearchPath + _findFontDirs(
                        '/usr/share/fonts/type1',
                        '/usr/share/fonts/Type1',
                        )
    TTFSearchPath = TTFSearchPath + _findFontDirs(
                        '/usr/share/fonts/truetype',
                        '/usr/share/fonts/TTF',
                        )



from reportlab.graphics.shapes import Drawing, Polygon, Line

def _getShaded(col,shd=None,shading=0.1):
    if shd is None:
        from reportlab.lib.colors import Blacker
        if col: shd = Blacker(col,1-shading)
    return shd

def _getLit(col,shd=None,lighting=0.1):
    if shd is None:
        from reportlab.lib.colors import Whiter
        if col: shd = Whiter(col,1-lighting)
    return shd


def _draw_3d_bar(G, x1, x2, y0, yhigh, xdepth, ydepth,
                fillColor=None, fillColorShaded=None,
                strokeColor=None, strokeWidth=1, shading=0.1):
    fillColorShaded = _getShaded(fillColor,None,shading)
    fillColorShadedTop = _getShaded(fillColor,None,shading/2.0)

    def _add_3d_bar(x1, x2, y1, y2, xoff, yoff,
                    G=G,strokeColor=strokeColor, strokeWidth=strokeWidth, fillColor=fillColor):
        G.add(Polygon((x1,y1, x1+xoff,y1+yoff, x2+xoff,y2+yoff, x2,y2),
            strokeWidth=strokeWidth, strokeColor=strokeColor, fillColor=fillColor,strokeLineJoin=1))

    usd = max(y0, yhigh)
    if xdepth or ydepth:
        if y0!=yhigh:   #non-zero height
            _add_3d_bar( x2, x2, y0, yhigh, xdepth, ydepth, fillColor=fillColorShaded) #side

        _add_3d_bar(x1, x2, usd, usd, xdepth, ydepth, fillColor=fillColorShadedTop)    #top

    G.add(Polygon((x1,y0,x2,y0,x2,yhigh,x1,yhigh),
        strokeColor=strokeColor, strokeWidth=strokeWidth, fillColor=fillColor,strokeLineJoin=1)) #front

    if xdepth or ydepth:
        G.add(Line( x1, usd, x2, usd, strokeWidth=strokeWidth, strokeColor=strokeColor or fillColorShaded))

class _YStrip:
    def __init__(self,y0,y1, slope, fillColor, fillColorShaded, shading=0.1):
        self.y0 = y0
        self.y1 = y1
        self.slope = slope
        self.fillColor = fillColor
        self.fillColorShaded = _getShaded(fillColor,fillColorShaded,shading)

def _ystrip_poly( x0, x1, y0, y1, xoff, yoff):
    return [x0,y0,x0+xoff,y0+yoff,x1+xoff,y1+yoff,x1,y1]


def _make_3d_line_info( G, x0, x1, y0, y1, z0, z1,
                    theta_x, theta_y,
                    fillColor, fillColorShaded=None, tileWidth=1,
                    strokeColor=None, strokeWidth=None, strokeDashArray=None,
                    shading=0.1):
    zwidth = abs(z1-z0)
    xdepth = zwidth*theta_x
    ydepth = zwidth*theta_y
    depth_slope  = xdepth==0 and 1e150 or -ydepth/float(xdepth)

    x = float(x1-x0)
    slope = x==0 and 1e150 or (y1-y0)/x

    c = slope>depth_slope and _getShaded(fillColor,fillColorShaded,shading) or fillColor
    zy0 = z0*theta_y
    zx0 = z0*theta_x

    tileStrokeWidth = 0.6
    if tileWidth is None:
        D = [(x1,y1)]
    else:
        T = ((y1-y0)**2+(x1-x0)**2)**0.5
        tileStrokeWidth *= tileWidth
        if T<tileWidth:
            D = [(x1,y1)]
        else:
            n = int(T/float(tileWidth))+1
            dx = float(x1-x0)/n
            dy = float(y1-y0)/n
            D = []
            a = D.append
            for i in range(1,n):
                a((x0+dx*i,y0+dy*i))

    a = G.add
    x_0 = x0+zx0
    y_0 = y0+zy0
    for x,y in D:
        x_1 = x+zx0
        y_1 = y+zy0
        P = Polygon(_ystrip_poly(x_0, x_1, y_0, y_1, xdepth, ydepth),
                    fillColor = c, strokeColor=c, strokeWidth=tileStrokeWidth)
        a((0,z0,z1,x_0,y_0,P))
        x_0 = x_1
        y_0 = y_1

from math import pi
_pi_2 = pi*0.5
_2pi = 2*pi
_180_pi=180./pi

def _2rad(angle):
    return angle/_180_pi

def mod_2pi(radians):
    radians = radians % _2pi
    if radians<-1e-6: radians += _2pi
    return radians

def _2deg(o):
    return o*_180_pi

def _360(a):
    a %= 360
    if a<-1e-6: a += 360
    return a

_ZERO = 1e-8
_ONE = 1-_ZERO
class _Segment:
    def __init__(self,s,i,data):
        S = data[s]
        x0 = S[i-1][0]
        y0 = S[i-1][1]
        x1 = S[i][0]
        y1 = S[i][1]
        if x1<x0:
            x0,y0,x1,y1 = x1,y1,x0,y0
        # (y-y0)*(x1-x0) = (y1-y0)*(x-x0)
        # (x1-x0)*y + (y0-y1)*x = y0*(x1-x0)+x0*(y0-y1)
        # a*y+b*x = c
        self.a = float(x1-x0)
        self.b = float(y1-y0)
        self.x0 = x0
        self.x1 = x1
        self.y0 = y0
        self.y1 = y1
        self.series = s
        self.i = i
        self.s = s

    def __str__(self):
        return '[(%s,%s),(%s,%s)]' % (self.x0,self.y0,self.x1,self.y1)

    __repr__ = __str__

    def intersect(self,o,I):
        '''try to find an intersection with _Segment o
        '''
        x0 = self.x0
        ox0 = o.x0
        assert x0<=ox0
        if ox0>self.x1: return 1
        if o.s==self.s and o.i in (self.i-1,self.i+1): return
        a = self.a
        b = self.b
        oa = o.a
        ob = o.b
        det = ob*a - oa*b
        if -1e-8<det<1e-8: return
        dx = x0 - ox0
        dy = self.y0 - o.y0
        u = (oa*dy - ob*dx)/det
        ou = (a*dy - b*dx)/det
        if u<0 or u>1 or ou<0 or ou>1: return
        x = x0 + u*a
        y = self.y0 + u*b
        if _ZERO<u<_ONE:
            t = self.s,self.i,x,y
            if t not in I: I.append(t)
        if _ZERO<ou<_ONE:
            t = o.s,o.i,x,y
            if t not in I:  I.append(t)

def _segKey(a):
    return (a.x0,a.x1,a.y0,a.y1,a.s,a.i)

def find_intersections(data,small=0):
    '''
    data is a sequence of series
    each series is a list of (x,y) coordinates
    where x & y are ints or floats

    find_intersections returns a sequence of 4-tuples
        i, j, x, y

    where i is a data index j is an insertion position for data[i]
    and x, y are coordinates of an intersection of series data[i]
    with some other series. If correctly implemented we get all such
    intersections. We don't count endpoint intersections and consider
    parallel lines as non intersecting (even when coincident).
    We ignore segments that have an estimated size less than small.
    '''

    #find all line segments
    S = []
    a = S.append
    for s in range(len(data)):
        ds = data[s]
        if not ds: continue
        n = len(ds)
        if n==1: continue
        for i in range(1,n):
            seg = _Segment(s,i,data)
            if seg.a+abs(seg.b)>=small: a(seg)
    S.sort(key=_segKey)
    I = []
    n = len(S)
    for i in range(0,n-1):
        s = S[i]
        for j in range(i+1,n):
            if s.intersect(S[j],I)==1: break
    I.sort()
    return I

if __name__=='__main__':
    from reportlab.graphics.shapes import Drawing
    from reportlab.lib.colors import lightgrey, pink
    D = Drawing(300,200)
    _draw_3d_bar(D, 10, 20, 10, 50, 5, 5, fillColor=lightgrey, strokeColor=pink)
    _draw_3d_bar(D, 30, 40, 10, 45, 5, 5, fillColor=lightgrey, strokeColor=pink)

    D.save(formats=['pdf'],outDir='.',fnRoot='_draw_3d_bar')

    print(find_intersections([[(0,0.5),(1,0.5),(0.5,0),(0.5,1)],[(.2666666667,0.4),(0.1,0.4),(0.1,0.2),(0,0),(1,1)],[(0,1),(0.4,0.1),(1,0.1)]]))
    print(find_intersections([[(0.1, 0.2), (0.1, 0.4)], [(0, 1), (0.4, 0.1)]]))
    print(find_intersections([[(0.2, 0.4), (0.1, 0.4)], [(0.1, 0.8), (0.4, 0.1)]]))
    print(find_intersections([[(0,0),(1,1)],[(0.4,0.1),(1,0.1)]]))
    print(find_intersections([[(0,0.5),(1,0.5),(0.5,0),(0.5,1)],[(0,0),(1,1)],[(0.1,0.8),(0.4,0.1),(1,0.1)]]))



#Copyright ReportLab Europe Ltd. 2000-2017
#see license.txt for license details
#history https://hg.reportlab.com/hg-public/reportlab/log/tip/src/reportlab/graphics/widgetbase.py
__version__='3.3.0'
__doc__='''Base class for user-defined graphical widgets'''

from reportlab.graphics import shapes
from reportlab import rl_config
from reportlab.lib import colors
from reportlab.lib.validators import *
from reportlab.lib.attrmap import *
from weakref import ref as weakref_ref

class PropHolder:
    '''Base for property holders'''

    _attrMap = None

    def verify(self):
        """If the _attrMap attribute is not None, this
        checks all expected attributes are present; no
        unwanted attributes are present; and (if a
        checking function is found) checks each
        attribute has a valid value.  Either succeeds
        or raises an informative exception.
        """

        if self._attrMap is not None:
            for key in self.__dict__.keys():
                if key[0] != '_':
                    msg = "Unexpected attribute %s found in %s" % (key, self)
                    assert key in self._attrMap, msg
            for attr, metavalue in self._attrMap.items():
                msg = "Missing attribute %s from %s" % (attr, self)
                assert hasattr(self, attr), msg
                value = getattr(self, attr)
                args = (value, attr, self.__class__.__name__)
                assert metavalue.validate(value), "Invalid value %s for attribute %s in class %s" % args

    if rl_config.shapeChecking:
        """This adds the ability to check every attribute assignment
        as it is made. It slows down shapes but is a big help when
        developing. It does not get defined if rl_config.shapeChecking = 0.
        """

        def __setattr__(self, name, value):
            """By default we verify.  This could be off
            in some parallel base classes."""
            validateSetattr(self,name,value)


    def getProperties(self,recur=1):
        """Returns a list of all properties which can be edited and
        which are not marked as private. This may include 'child
        widgets' or 'primitive shapes'.  You are free to override
        this and provide alternative implementations; the default
        one simply returns everything without a leading underscore.
        """

        from reportlab.lib.validators import isValidChild

        # TODO when we need it, but not before -
        # expose sequence contents?

        props = {}
        for name in self.__dict__.keys():
            if name[0:1] != '_':
                component = getattr(self, name)

                if recur and isValidChild(component):
                    # child object, get its properties too
                    childProps = component.getProperties(recur=recur)
                    for childKey, childValue in childProps.items():
                        #key might be something indexed like '[2].fillColor'
                        #or simple like 'fillColor'; in the former case we
                        #don't need a '.' between me and my child.
                        if childKey[0] == '[':
                            props['%s%s' % (name, childKey)] = childValue
                        else:
                            props['%s.%s' % (name, childKey)] = childValue
                else:
                    props[name] = component

        return props


    def setProperties(self, propDict):
        """Permits bulk setting of properties.  These may include
        child objects e.g. "chart.legend.width = 200".

        All assignments will be validated by the object as if they
        were set individually in python code.

        All properties of a top-level object are guaranteed to be
        set before any of the children, which may be helpful to
        widget designers.
        """

        childPropDicts = {}
        for name, value in propDict.items():
            parts = name.split('.', 1)
            if len(parts) == 1:
                #simple attribute, set it now
                setattr(self, name, value)
            else:
                (childName, remains) = parts
                try:
                    childPropDicts[childName][remains] = value
                except KeyError:
                    childPropDicts[childName] = {remains: value}

        # now assign to children
        for childName, childPropDict in childPropDicts.items():
            child = getattr(self, childName)
            child.setProperties(childPropDict)


    def dumpProperties(self, prefix=""):
        """Convenience. Lists them on standard output.  You
        may provide a prefix - mostly helps to generate code
        samples for documentation.
        """

        propList = list(self.getProperties().items())
        propList.sort()
        if prefix:
            prefix = prefix + '.'
        for (name, value) in propList:
            print('%s%s = %s' % (prefix, name, value))


class Widget(PropHolder, shapes.UserNode):
    """Base for all user-defined widgets.  Keep as simple as possible. Does
    not inherit from Shape so that we can rewrite shapes without breaking
    widgets and vice versa."""

    def _setKeywords(self,**kw):
        for k,v in kw.items():
            if k not in self.__dict__:
                setattr(self,k,v)

    def draw(self):
        msg = "draw() must be implemented for each Widget!"
        raise NotImplementedError(msg)

    def demo(self):
        msg = "demo() must be implemented for each Widget!"
        raise NotImplementedError(msg)

    def provideNode(self):
        return self.draw()

    def getBounds(self):
        "Return outer boundary as x1,y1,x2,y2.  Can be overridden for efficiency"
        return self.draw().getBounds()

class ScaleWidget(Widget):
    '''Contents with a scale and offset''' 
    _attrMap = AttrMap(
        x = AttrMapValue(isNumber,desc="x offset"),
        y = AttrMapValue(isNumber,desc="y offset"),
        scale = AttrMapValue(isNumber,desc="scale"),
        contents = AttrMapValue(None,desc="Contained drawable elements"),
        )
    def __init__(self,x=0,y=0,scale=1.0,contents=None):
        self.x = x
        self.y = y
        if not contents: contents=[]
        elif not isinstance(contents,(tuple,list)):
            contents = (contents,)
        self.contents = list(contents)
        self.scale = scale
    
    def draw(self):
        return shapes.Group(transform=(self.scale,0,0,self.scale,self.x,self.y),*self.contents)

_ItemWrapper={}

class CloneMixin:
    def clone(self,**kwds):
        n = self.__class__()
        n.__dict__.clear()
        n.__dict__.update(self.__dict__)
        if kwds: n.__dict__.update(kwds)
        return n

class TypedPropertyCollection(PropHolder):
    """A container with properties for objects of the same kind.

    This makes it easy to create lists of objects. You initialize
    it with a class of what it is to contain, and that is all you
    can add to it.  You can assign properties to the collection
    as a whole, or to a numeric index within it; if so it creates
    a new child object to hold that data.

    So:
        wedges = TypedPropertyCollection(WedgeProperties)
        wedges.strokeWidth = 2                # applies to all
        wedges.strokeColor = colors.red       # applies to all
        wedges[3].strokeColor = colors.blue   # only to one

    The last line should be taken as a prescription of how to
    create wedge no. 3 if one is needed; no error is raised if
    there are only two data points.

    We try and make sensible use of tuple indices.
        line[(3,x)] is backed by line[(3,)] == line[3] & line
    """

    def __init__(self, exampleClass, **kwds):
        #give it same validation rules as what it holds
        self.__dict__['_value'] = exampleClass(**kwds)
        self.__dict__['_children'] = {}

    def wKlassFactory(self,Klass):
        class WKlass(Klass,CloneMixin):
            def __getattr__(self,name):
                try:
                    return self.__class__.__bases__[0].__getattr__(self,name)
                except:
                    parent = self.parent
                    c = parent._children
                    x = self.__propholder_index__
                    while x:
                        if x in c:
                            return getattr(c[x],name)
                        x = x[:-1]
                    return getattr(parent,name)
            @property
            def parent(self):
                return self.__propholder_parent__()
        return WKlass

    def __getitem__(self, x):
        x = tuple(x) if isinstance(x,(tuple,list)) else (x,)
        try:
            return self._children[x]
        except KeyError:
            Klass = self._value.__class__
            if Klass in _ItemWrapper:
                WKlass = _ItemWrapper[Klass]
            else:
                _ItemWrapper[Klass] = WKlass = self.wKlassFactory(Klass)

            child = WKlass()
            
            for i in filter(lambda x,K=list(child.__dict__.keys()): x in K,list(child._attrMap.keys())):
                del child.__dict__[i]
            child.__dict__.update(dict(
                                    __propholder_parent__ = weakref_ref(self),
                                    __propholder_index__ = x[:-1])
                                    )

            self._children[x] = child
            return child

    def __contains__(self,key):
        return (tuple(key) if isinstance(key,(tuple,list)) else (key,)) in self._children

    def __setitem__(self, key, value):
        assert isinstance(value, self._value.__class__), (
            "This collection can only hold objects of type %s" % self._value.__class__.__name__)

    def __len__(self):
        return len(list(self._children.keys()))

    def getProperties(self,recur=1):
        # return any children which are defined and whatever
        # differs from the parent
        props = {}

        for key, value in self._value.getProperties(recur=recur).items():
            props['%s' % key] = value

        for idx in self._children.keys():
            childProps = self._children[idx].getProperties(recur=recur)
            for key, value in childProps.items():
                if not hasattr(self,key) or getattr(self, key)!=value:
                    newKey = '[%s].%s' % (idx if len(idx)>1 else idx[0], key)
                    props[newKey] = value
        return props

    def setVector(self,**kw):
        for name, value in kw.items():
            for i, v in enumerate(value):
                setattr(self[i],name,v)

    def __getattr__(self,name):
        return getattr(self._value,name)

    def __setattr__(self,name,value):
        return setattr(self._value,name,value)

    def checkAttr(self, key, a, default=None):
        return getattr(self[key], a, default) if key in self else default

def tpcGetItem(obj,x):
    '''return obj if it's not a TypedPropertyCollection else obj[x]'''
    return obj[x] if isinstance(obj,TypedPropertyCollection) else obj

def isWKlass(obj):
    if not hasattr(obj,'__propholder_parent__'): return
    ph = obj.__propholder_parent__
    if not isinstance(ph,weakref_ref): return
    return isinstance(ph(),TypedPropertyCollection)

## No longer needed!
class StyleProperties(PropHolder):
    """A container class for attributes used in charts and legends.

    Attributes contained can be those for any graphical element
    (shape?) in the ReportLab graphics package. The idea for this
    container class is to be useful in combination with legends
    and/or the individual appearance of data series in charts.

    A legend could be as simple as a wrapper around a list of style
    properties, where the 'desc' attribute contains a descriptive
    string and the rest could be used by the legend e.g. to draw
    something like a color swatch. The graphical presentation of
    the legend would be its own business, though.

    A chart could be inspecting a legend or, more directly, a list
    of style properties to pick individual attributes that it knows
    about in order to render a particular row of the data. A bar
    chart e.g. could simply use 'strokeColor' and 'fillColor' for
    drawing the bars while a line chart could also use additional
    ones like strokeWidth.
    """

    _attrMap = AttrMap(
        strokeWidth = AttrMapValue(isNumber,desc='width of the stroke line'),
        strokeLineCap = AttrMapValue(isNumber,desc='Line cap 0=butt, 1=round & 2=square',advancedUsage=1),
        strokeLineJoin = AttrMapValue(isNumber,desc='Line join 0=miter, 1=round & 2=bevel',advancedUsage=1),
        strokeMiterLimit = AttrMapValue(None,desc='miter limit control miter line joins',advancedUsage=1),
        strokeDashArray = AttrMapValue(isListOfNumbersOrNone,desc='dashing patterns e.g. (1,3)'),
        strokeOpacity = AttrMapValue(isNumber,desc='level of transparency (alpha) accepts values between 0..1',advancedUsage=1),
        strokeColor = AttrMapValue(isColorOrNone,desc='the color of the stroke'),
        fillColor = AttrMapValue(isColorOrNone,desc='the filling color'),
        desc = AttrMapValue(isString),
        )

    def __init__(self, **kwargs):
        "Initialize with attributes if any."

        for k, v in kwargs.items():
            setattr(self, k, v)


    def __setattr__(self, name, value):
        "Verify attribute name and value, before setting it."
        validateSetattr(self,name,value)


class TwoCircles(Widget):
    def __init__(self):
        self.leftCircle = shapes.Circle(100,100,20, fillColor=colors.red)
        self.rightCircle = shapes.Circle(300,100,20, fillColor=colors.red)

    def draw(self):
        return shapes.Group(self.leftCircle, self.rightCircle)


class Face(Widget):
    """This draws a face with two eyes.

    It exposes a couple of properties
    to configure itself and hides all other details.
    """

    _attrMap = AttrMap(
        x = AttrMapValue(isNumber),
        y = AttrMapValue(isNumber),
        size = AttrMapValue(isNumber),
        skinColor = AttrMapValue(isColorOrNone),
        eyeColor = AttrMapValue(isColorOrNone),
        mood = AttrMapValue(OneOf('happy','sad','ok')),
        )

    def __init__(self):
        self.x = 10
        self.y = 10
        self.size = 80
        self.skinColor = None
        self.eyeColor = colors.blue
        self.mood = 'happy'

    def demo(self):
        pass

    def draw(self):
        s = self.size  # abbreviate as we will use this a lot
        g = shapes.Group()
        g.transform = [1,0,0,1,self.x, self.y]

        # background
        g.add(shapes.Circle(s * 0.5, s * 0.5, s * 0.5, fillColor=self.skinColor))

        # left eye
        g.add(shapes.Circle(s * 0.35, s * 0.65, s * 0.1, fillColor=colors.white))
        g.add(shapes.Circle(s * 0.35, s * 0.65, s * 0.05, fillColor=self.eyeColor))

        # right eye
        g.add(shapes.Circle(s * 0.65, s * 0.65, s * 0.1, fillColor=colors.white))
        g.add(shapes.Circle(s * 0.65, s * 0.65, s * 0.05, fillColor=self.eyeColor))

        # nose
        g.add(shapes.Polygon(
            points=[s * 0.5, s * 0.6, s * 0.4, s * 0.3, s * 0.6, s * 0.3],
            fillColor=None))

        # mouth
        if self.mood == 'happy':
            offset = -0.05
        elif self.mood == 'sad':
            offset = +0.05
        else:
            offset = 0

        g.add(shapes.Polygon(
            points = [
                s * 0.3, s * 0.2, #left of mouth
                s * 0.7, s * 0.2, #right of mouth
                s * 0.6, s * (0.2 + offset), # the bit going up or down
                s * 0.4, s * (0.2 + offset) # the bit going up or down
                ],
            fillColor = colors.pink,
            strokeColor = colors.red,
            strokeWidth = s * 0.03
            ))

        return g


class TwoFaces(Widget):
    def __init__(self):
        self.faceOne = Face()
        self.faceOne.mood = "happy"
        self.faceTwo = Face()
        self.faceTwo.x = 100
        self.faceTwo.mood = "sad"

    def draw(self):
        """Just return a group"""
        return shapes.Group(self.faceOne, self.faceTwo)

    def demo(self):
        """The default case already looks good enough,
        no implementation needed here"""
        pass

class Sizer(Widget):
    "Container to show size of all enclosed objects"

    _attrMap = AttrMap(BASE=shapes.SolidShape,
        contents = AttrMapValue(isListOfShapes,desc="Contained drawable elements"),
        )
    def __init__(self, *elements):
        self.contents = []
        self.fillColor = colors.cyan
        self.strokeColor = colors.magenta

        for elem in elements:
            self.add(elem)

    def _addNamedNode(self,name,node):
        'if name is not None add an attribute pointing to node and add to the attrMap'
        if name:
            if name not in list(self._attrMap.keys()):
                self._attrMap[name] = AttrMapValue(isValidChild)
            setattr(self, name, node)

    def add(self, node, name=None):
        """Appends non-None child node to the 'contents' attribute. In addition,
        if a name is provided, it is subsequently accessible by name
        """
        # propagates properties down
        if node is not None:
            assert isValidChild(node), "Can only add Shape or UserNode objects to a Group"
            self.contents.append(node)
            self._addNamedNode(name,node)

    def getBounds(self):
        # get bounds of each object
        if self.contents:
            b = []
            for elem in self.contents:
                b.append(elem.getBounds())
            return shapes.getRectsBounds(b)
        else:
            return (0,0,0,0)

    def draw(self):
        g = shapes.Group()
        (x1, y1, x2, y2) = self.getBounds()
        r = shapes.Rect(
            x = x1,
            y = y1,
            width = x2-x1,
            height = y2-y1,
            fillColor = self.fillColor,
            strokeColor = self.strokeColor
            )
        g.add(r)
        for elem in self.contents:
            g.add(elem)
        return g

class CandleStickProperties(PropHolder):
    _attrMap = AttrMap(
        strokeWidth = AttrMapValue(isNumber, desc='Width of a line.'),
        strokeColor = AttrMapValue(isColorOrNone, desc='Color of a line or border.'),
        strokeDashArray = AttrMapValue(isListOfNumbersOrNone, desc='Dash array of a line.'),
        crossWidth = AttrMapValue(isNumberOrNone,desc="cross line width",advancedUsage=1),
        crossLo = AttrMapValue(isNumberOrNone,desc="cross line low value",advancedUsage=1),
        crossHi = AttrMapValue(isNumberOrNone,desc="cross line high value",advancedUsage=1),
        boxWidth = AttrMapValue(isNumberOrNone,desc="width of the box part",advancedUsage=1),
        boxFillColor = AttrMapValue(isColorOrNone, desc='fill color of box'),
        boxStrokeColor = AttrMapValue(NotSetOr(isColorOrNone), desc='stroke color of box'),
        boxStrokeDashArray = AttrMapValue(NotSetOr(isListOfNumbersOrNone), desc='Dash array of the box.'),
        boxStrokeWidth = AttrMapValue(NotSetOr(isNumber), desc='Width of the box lines.'),
        boxLo = AttrMapValue(isNumberOrNone,desc="low value of the box",advancedUsage=1),
        boxMid = AttrMapValue(isNumberOrNone,desc="middle box line value",advancedUsage=1),
        boxHi = AttrMapValue(isNumberOrNone,desc="high value of the box",advancedUsage=1),
        boxSides = AttrMapValue(isBoolean,desc="whether to show box sides",advancedUsage=1),
        position = AttrMapValue(isNumberOrNone,desc="position of the candle",advancedUsage=1),
        chart = AttrMapValue(None,desc="our chart",advancedUsage=1),
        candleKind = AttrMapValue(OneOf('vertical','horizontal'),desc="candle direction",advancedUsage=1),
        axes = AttrMapValue(SequenceOf(isString,emptyOK=0,lo=2,hi=2),desc="candle direction",advancedUsage=1),
        )

    def __init__(self,**kwds):
        self.strokeWidth = kwds.pop('strokeWidth',1)
        self.strokeColor = kwds.pop('strokeColor',colors.black)
        self.strokeDashArray = kwds.pop('strokeDashArray',None)
        self.crossWidth = kwds.pop('crossWidth',5)
        self.crossLo = kwds.pop('crossLo',None)
        self.crossHi = kwds.pop('crossHi',None)
        self.boxWidth = kwds.pop('boxWidth',None)
        self.boxFillColor = kwds.pop('boxFillColor',None)
        self.boxStrokeColor =kwds.pop('boxStrokeColor',NotSetOr._not_set) 
        self.boxStrokeWidth =kwds.pop('boxStrokeWidth',NotSetOr._not_set) 
        self.boxStrokeDashArray =kwds.pop('boxStrokeDashArray',NotSetOr._not_set) 
        self.boxLo = kwds.pop('boxLo',None)
        self.boxMid = kwds.pop('boxMid',None)
        self.boxHi = kwds.pop('boxHi',None)
        self.boxSides = kwds.pop('boxSides',True)
        self.position = kwds.pop('position',None)
        self.candleKind = kwds.pop('candleKind','vertical')
        self.axes = kwds.pop('axes',['categoryAxis','valueAxis'])
        chart = kwds.pop('chart',None)
        self.chart = weakref_ref(chart) if chart else (lambda:None)

    def __call__(self,_x,_y,_size,_color):
        '''the symbol interface'''
        chart = self.chart()
        xA = getattr(chart,self.axes[0])
        _xScale = getattr(xA,'midScale',None)
        if not _xScale: _xScale = getattr(xA,'scale')
        xScale = lambda x: _xScale(x) if x is not None else None
        yA = getattr(chart,self.axes[1])
        _yScale = getattr(yA,'midScale',None)
        if not _yScale: _yScale = getattr(yA,'scale')
        yScale = lambda x: _yScale(x) if x is not None else None
        G = shapes.Group().add
        strokeWidth = self.strokeWidth
        strokeColor = self.strokeColor
        strokeDashArray = self.strokeDashArray
        crossWidth = self.crossWidth
        crossLo = yScale(self.crossLo)
        crossHi = yScale(self.crossHi)
        boxWidth = self.boxWidth
        boxFillColor = self.boxFillColor
        boxStrokeColor = NotSetOr.conditionalValue(self.boxStrokeColor,strokeColor)
        boxStrokeWidth = NotSetOr.conditionalValue(self.boxStrokeWidth,strokeWidth)
        boxStrokeDashArray = NotSetOr.conditionalValue(self.boxStrokeDashArray,strokeDashArray)
        boxLo = yScale(self.boxLo)
        boxMid = yScale(self.boxMid)
        boxHi = yScale(self.boxHi)
        position = xScale(self.position)
        candleKind = self.candleKind
        haveBox = None not in (boxWidth,boxLo,boxHi)
        haveLine = None not in (crossLo,crossHi)
        def aLine(x0,y0,x1,y1):
            if candleKind!='vertical':
                x0,y0 = y0,x0
                x1,y1 = y1,x1
            G(shapes.Line(x0,y0,x1,y1,strokeWidth=strokeWidth,strokeColor=strokeColor,strokeDashArray=strokeDashArray))
        if haveBox:
            boxLo, boxHi = min(boxLo,boxHi), max(boxLo,boxHi)
        if haveLine:
            crossLo, crossHi = min(crossLo,crossHi), max(crossLo,crossHi)
            if not haveBox or crossLo>=boxHi or crossHi<=boxLo:
                aLine(position,crossLo,position,crossHi)
                if crossWidth is not None:
                    aLine(position-crossWidth*0.5,crossLo,position+crossWidth*0.5,crossLo)
                    aLine(position-crossWidth*0.5,crossHi,position+crossWidth*0.5,crossHi)
            elif haveBox:
                if crossLo<boxLo:
                    aLine(position,crossLo,position,boxLo)
                    aLine(position-crossWidth*0.5,crossLo,position+crossWidth*0.5,crossLo)
                if crossHi>boxHi:
                    aLine(position,boxHi,position,crossHi)
                    aLine(position-crossWidth*0.5,crossHi,position+crossWidth*0.5,crossHi)
        if haveBox:
            x = position - boxWidth*0.5
            y = boxLo
            h = boxHi - boxLo
            w = boxWidth
            if candleKind!='vertical':
                x, y, w, h = y, x, h, w
            G(shapes.Rect(x,y,w,h,strokeColor=boxStrokeColor if self.boxSides else None,strokeWidth=boxStrokeWidth,strokeDashArray=boxStrokeDashArray,fillColor=boxFillColor))
            if not self.boxSides:
                aLine(position-0.5*boxWidth,boxHi,position+0.5*boxWidth,boxHi)
                aLine(position-0.5*boxWidth,boxLo,position+0.5*boxWidth,boxLo)

            if boxMid is not None:
                aLine(position-0.5*boxWidth,boxMid,position+0.5*boxWidth,boxMid)
        return G.__self__

def CandleSticks(**kwds):
    return TypedPropertyCollection(CandleStickProperties,**kwds)

def test():
    from reportlab.graphics.charts.piecharts import WedgeProperties
    wedges = TypedPropertyCollection(WedgeProperties)
    wedges.fillColor = colors.red
    wedges.setVector(fillColor=(colors.blue,colors.green,colors.white))
    print(len(_ItemWrapper))

    d = shapes.Drawing(400, 200)
    tc = TwoCircles()
    d.add(tc)
    from reportlab.graphics import renderPDF
    renderPDF.drawToFile(d, 'sample_widget.pdf', 'A Sample Widget')
    print('saved sample_widget.pdf')

    d = shapes.Drawing(400, 200)
    f = Face()
    f.skinColor = colors.yellow
    f.mood = "sad"
    d.add(f, name='theFace')
    print('drawing 1 properties:')
    d.dumpProperties()
    renderPDF.drawToFile(d, 'face.pdf', 'A Sample Widget')
    print('saved face.pdf')

    d2 = d.expandUserNodes()
    renderPDF.drawToFile(d2, 'face_copy.pdf', 'An expanded drawing')
    print('saved face_copy.pdf')
    print('drawing 2 properties:')
    d2.dumpProperties()


if __name__=='__main__':
    test()
